# Copyright contributors to the TSFM project
#
"""Primary entry point for inference services"""

import logging

import starlette.status as status
from fastapi import FastAPI
from fastapi.responses import RedirectResponse
from prometheus_client import CollectorRegistry, make_asgi_app, multiprocess

from . import (
    TSFM_MODEL_DIR,
    TSFM_PYTHON_LOGGING_FORMAT,
    TSFM_PYTHON_LOGGING_LEVEL,
)
from .constants import API_VERSION
from .inference import InferenceRuntime


logging.basicConfig(
    format=TSFM_PYTHON_LOGGING_FORMAT,
    level=TSFM_PYTHON_LOGGING_LEVEL,
)

logging.info(f"Using TSFM_MODEL_DIR {TSFM_MODEL_DIR}")

app = FastAPI(
    title="FM for Time Series API",
    version=API_VERSION,
    description="This FastAPI application provides service endpoints for performing inference tasks on TSFM HF models.",
)


# ############# PROMETHEUS METRICS ##############
# Using multiprocess collector for registry
def make_metrics_app():
    registry = CollectorRegistry()
    multiprocess.MultiProcessCollector(registry)
    return make_asgi_app(registry=registry)


metrics_app = make_metrics_app()
app.mount("/metrics", metrics_app)
# ##############################################

ir = InferenceRuntime()
ir.add_routes(app)


# @app.get("/routes")
# def test_route():
#    return {"routes": [route.path for route in app.routes]}


@app.get("/healthcheck")
def healthcheck():
    return {"message": "Server is healthy!"}


@app.get("/")
def root():
    return RedirectResponse(url="/docs", status_code=status.HTTP_302_FOUND)


if __name__ == "__main__":
    # Third Party
    import uvicorn

    # Run the FastAPI application on the local host and port 7860
    # CMD["uvicorn", "app:app", "--host", "127.0.0.1", "--port", "7860", "--reload"]
    uvicorn.run(app, host="127.0.0.1", port=8000)



# Copyright contributors to the TSFM project
#

"""Tests the time series classification pipeline"""

import pytest

from tsfm_public.models.tspulse import TSPulseConfig, TSPulseForClassification
from tsfm_public.toolkit.time_series_classification_pipeline import TimeSeriesClassificationPipeline
from tsfm_public.toolkit.time_series_classification_preprocessor import TimeSeriesClassificationPreprocessor


@pytest.fixture(scope="module")
def tspulse_model():
    conf = TSPulseConfig(
        context_length=512,
        d_model=24,
        decoder_d_model=24,
        num_patches=128,
        patch_stride=8,
        fft_time_add_forecasting_pt_loss=False,
        num_input_channels=2,  # ts_data_nested has 2 input_columns
        patch_length=8,
        mask_block_length=2,
        mode="common_channel",
        gated_attn=True,
        use_positional_encoding=False,
        self_attn=False,
        head_aggregation="max_pool",
        fuse_fft=True,
        use_learnable_mask_token=True,
        prediction_length=4,
        channel_mix_init="identity",
        loss="cross_entropy",
        num_targets=4,  # set to 4 classes as in the ts_data_nested
    )
    model = TSPulseForClassification(conf)
    return model


def test_time_series_classification_pipeline_defaults(tspulse_model, ts_data_nested):
    df = ts_data_nested.copy()

    tsp = TimeSeriesClassificationPreprocessor(
        timestamp_column="time_date",
        input_columns=["val", "val2"],
        label_column="label",
        id_columns=[
            "id",
        ],
    )
    tsp.train(df)

    pipe = TimeSeriesClassificationPipeline(tspulse_model, feature_extractor=tsp, device="cpu")

    out_df = pipe(df)

    assert df.shape[0] == out_df.shape[0]
    assert all(t in out_df.columns for t in ["val", "val2"])
    assert "label_prediction" in out_df.columns


def test_time_series_classification_pipeline_predictions(tspulse_model, ts_data_nested):
    df = ts_data_nested.copy()

    tsp = TimeSeriesClassificationPreprocessor(
        timestamp_column="time_date",
        input_columns=["val", "val2"],
        label_column="label",
        id_columns=[
            "id",
        ],
    )
    tsp.train(df)

    pipe = TimeSeriesClassificationPipeline(tspulse_model, feature_extractor=tsp, device="cpu")

    out_df = pipe(df)

    assert out_df["label_prediction"].isin(out_df["label"]).all()


def test_time_series_classification_pipeline_inputs_integrity(tspulse_model, ts_data_nested):
    df = ts_data_nested.copy()

    tsp = TimeSeriesClassificationPreprocessor(
        timestamp_column="time_date",
        input_columns=["val", "val2"],
        label_column="label",
        id_columns=[
            "id",
        ],
        scaling=True,
    )
    tsp.train(df)

    pipe = TimeSeriesClassificationPipeline(tspulse_model, feature_extractor=tsp, device="cpu")

    out_df = pipe(df)

    assert ((out_df["val"] == df["val"]) & (out_df["val2"] == df["val2"])).all()



"""Base serivce handler"""

import enum
import importlib
import logging
from pathlib import Path
from typing import Optional, Tuple, Union

import pandas as pd

from tsfm_public.toolkit.tsfm_config import TSFMConfig

from .inference_payloads import (
    BaseMetadataInput,
    BaseParameters,
)


LOGGER = logging.getLogger(__file__)


class HandlerFunction(enum.Enum):
    """`Enum` for the different functions for which we use handlers."""

    INFERENCE = "inference"
    TUNING = "tuning"


class ServiceHandler:
    """Abstraction to enable serving of various models.

    Args:
        implementation (object): An instantiated class that implements methods needed to perform the required handler
            functions.
    """

    def __init__(self, implementation: object):
        # , model_id: str, model_path: Union[str, Path], handler_config: TSFMConfig, implementation: type):
        # self.model_id = model_id
        # self.model_path = model_path
        # self.handler_config = handler_config
        # self.prepared = False
        self.implementation = implementation

    @classmethod
    def load(
        cls,
        model_id: str,
        model_path: Union[str, Path],
        handler_function: str,
    ) -> Tuple["ServiceHandler", None] | Tuple[None, Exception]:
        """Load the handler_config -- the tsfm service config for this model, returning the proper
        handler to use the model.

        model_path is expected to point to a folder containing the tsfm_config.json file. This can be a local folder
        or with a model on the HuggingFace Hub.

        Args:
            model_id (str): A string identifier for the model.
            model_path (Union[str, Path]): The full path to the model, can be a local path or a HuggingFace Hub path.
            handler_function (str): The type of handler, currently supported handlers are defined in the HandlerFunction
                enum.

        """

        handler_config_path = Path(model_path) if isinstance(model_path, str) else model_path

        try:
            config = TSFMConfig.from_pretrained(handler_config_path)
        except (FileNotFoundError, OSError):
            LOGGER.info("TSFM Config file not found.")
            config = TSFMConfig()
        try:
            handler_class = get_service_handler_class(config, handler_function)

            # could validate handler_class here

            return cls(
                implementation=handler_class(
                    model_id=model_id,
                    model_path=model_path,
                    handler_config=config,
                )
            ), None
        except Exception as e:
            return None, e

    def prepare(
        self,
        data: pd.DataFrame,
        schema: Optional[BaseMetadataInput] = None,
        parameters: Optional[BaseParameters] = None,
        **kwargs,
    ) -> Tuple["ServiceHandler", None] | Tuple[None, Exception]:
        """Prepare the wrapper by loading all the components needed to use the model.

        The actual preparation is done in the `_prepare()` method -- which should be overridden for a
        particular model implementation. This is separate from `load()` above because we may need to know
        certain details about the model (learned from the handler config) before we can effectively load
        and configure the model artifacts.

        Args:
            data (pd.DataFrame): A pandas dataframe containing historical data.
            schema (Optional[BaseMetadataInput], optional): Service request schema. Defaults to None.
            parameters (Optional[BaseParameters], optional): Service requst parameters. Defaults to None.

        Returns:
            Tuple(ServiceHandler, None) or Tuple(None, Exception): In the first case, a tuple containing
                a prepared service handler is returned. The prepared service handler contains all the
                necessary artifacts for performing subsequent inference or training tasks. In the second
                case, the tuple contains an error object.
        """
        try:
            self.implementation.prepare(data=data, schema=schema, parameters=parameters, **kwargs)
            self.prepared = True
            return self, None
        except Exception as e:
            return self, e

    @property
    def handler_config(self):
        if self.implementation is not None:
            return self.implementation.handler_config
        else:
            return None


def get_service_handler_class(
    config: TSFMConfig, handler_function: str = HandlerFunction.INFERENCE.value
) -> "ServiceHandler":
    if handler_function == HandlerFunction.INFERENCE.value:
        handler_module_path_identifier = "inference_handler_module_path"
        handler_class_name_identifier = "inference_handler_class_name"
    elif handler_function == HandlerFunction.TUNING.value:
        handler_module_path_identifier = "tuning_handler_module_path"
        handler_class_name_identifier = "tuning_handler_class_name"
    else:
        raise ValueError(f"Unknown handler_function `{handler_function}`")

    if getattr(config, handler_module_path_identifier, None) and getattr(config, handler_class_name_identifier, None):
        module = importlib.import_module(getattr(config, handler_module_path_identifier))
        my_class = getattr(module, getattr(config, handler_class_name_identifier))

    elif handler_function == HandlerFunction.INFERENCE.value:
        # Default to forecasting task, inference
        from .tsfm_inference_handler import TSFMForecastingInferenceHandler

        my_class = TSFMForecastingInferenceHandler
    elif handler_function == HandlerFunction.TUNING.value:
        # Default to forecasting task, tuning
        from .tsfm_tuning_handler import TSFMForecastingTuningHandler

        my_class = TSFMForecastingTuningHandler

    return my_class



# Copyright contributors to the TSFM project
#
"""Recursive prediction model wrapper"""

import math
from dataclasses import dataclass
from typing import Optional

import torch
import torch.nn as nn
from transformers.configuration_utils import PretrainedConfig
from transformers.modeling_utils import PreTrainedModel
from transformers.utils import (
    ModelOutput,
)


@dataclass
class RecursivePredictorOutput(ModelOutput):
    """
    Output type of [`RecursivePredictorOutput`].

    Args:
        prediction_outputs (`torch.FloatTensor` of shape `(batch_size, prediction_length, num_input_channels)`):
            Prediction output from the prediction head.
        loss (*optional*, returned when `y` is provided, `torch.FloatTensor` of shape `()`):
            Total loss.
    """

    loss: Optional[torch.FloatTensor] = None
    prediction_outputs: torch.FloatTensor = None
    backbone_hidden_state: torch.FloatTensor = None


class RecursivePredictorConfig(PretrainedConfig):
    model_type = "recursivepredictor"
    """
    RecursivePredictorConfig

    Args:
        model (PreTrainedModel): Model to load for recursive predictions
        requested_prediction_length (int): Total prediction length
        model_prediction_length (int): prediction length of the model
        loss (str): loss to report
    """

    def __init__(
        self,
        model: PreTrainedModel,
        requested_prediction_length: int,
        model_prediction_length: int,
        loss: str,
        **kwargs,
    ):
        self.model = model
        self.requested_prediction_length = requested_prediction_length
        self.model_prediction_length = model_prediction_length
        self.loss = loss
        super().__init__(**kwargs)


class RecursivePredictorPreTrainedModel(PreTrainedModel):
    # Weight initialization
    config_class = RecursivePredictorConfig
    base_model_prefix = "model"
    main_input_name = "past_values"
    supports_gradient_checkpointing = False

    def _init_weights(self, module):
        """Initialize weights"""
        pass


class RecursivePredictor(RecursivePredictorPreTrainedModel):
    def __init__(self, config: RecursivePredictorConfig):
        super().__init__(config)
        self.model = config.model
        self.requested_prediction_length = config.requested_prediction_length
        self.model_prediction_length = config.model_prediction_length
        self.use_return_dict = config.use_return_dict
        if config.loss == "mse":
            self.loss = nn.MSELoss(reduction="mean")
        elif config.loss == "mae":
            self.loss = nn.L1Loss(reduction="mean")
        else:
            raise ValueError("Invalid loss function: Allowed values: mse and mae")

    def forward(
        self,
        past_values: torch.Tensor,
        future_values: Optional[torch.Tensor] = None,
        past_observed_mask: Optional[torch.Tensor] = None,
        future_observed_mask: Optional[torch.Tensor] = None,
        output_hidden_states: Optional[bool] = False,
        return_loss: bool = True,
        return_dict: Optional[bool] = None,
        freq_token: Optional[torch.Tensor] = None,
        static_categorical_values: Optional[torch.Tensor] = None,
    ) -> RecursivePredictorOutput:
        """
        Predict future points given an input sequence, using a recursive strategy.

        Assumptions: The model passed as part of the initialization, should support the following:
         - the signature of the forward method should allow the following arguments: past_values,
         future_vales, past_observed_mask, future_observed_mask, freq_token.
         - the model should have a config attribute prediction_channel_indices which indicates the
         indices in the past and future_value tensors which correspond to the channels we wish to predict
         - if future_values is provided, it must be of shape compatible with the requested_prediction_length

        Args:
            past_values (torch.Tensor): Input sequence of shape (batch_size, sequence_length, num_channels).
            requested_prediction_length (int): Number of future points to predict beyond the input sequence.

        Returns:
            predicted_sequence (torch.Tensor): Predicted sequence of shape (batch_size,
            requested_prediction_length, num_channels).
        """
        return_dict = return_dict if return_dict is not None else self.use_return_dict

        total_runs = math.ceil(self.requested_prediction_length / self.model_prediction_length)

        predicted_sequence = past_values.clone()  # Initialize predicted sequence with input sequence

        model_prediction_length = self.model_prediction_length

        this_past = past_values.clone()
        this_past_observed_mask = past_observed_mask.clone() if past_observed_mask is not None else None

        predicted_sequence = None

        prediction_channel_indices = self.model.config.prediction_channel_indices

        for i in range(total_runs):
            # index into the right part of the future
            future_start_idx = i * model_prediction_length
            future_end_idx = (i + 1) * model_prediction_length

            this_future = future_values[:, future_start_idx:future_end_idx] if future_values is not None else None
            this_future_observed_mask = (
                future_observed_mask[:, future_start_idx:future_end_idx] if future_observed_mask is not None else None
            )

            # predict and concatenate results
            next_point = self.model(
                past_values=this_past,
                future_values=this_future,
                past_observed_mask=this_past_observed_mask,
                future_observed_mask=this_future_observed_mask,
                freq_token=freq_token,
                static_categorical_values=static_categorical_values,
            )

            next_point = next_point["prediction_outputs"]
            predicted_sequence = (
                torch.cat((predicted_sequence, next_point), dim=1) if predicted_sequence is not None else next_point
            )

            # create the new part of the past, using the current future and new predictions
            if this_future is not None and prediction_channel_indices is not None:
                new_past = this_future.clone()
                new_past[:, :, self.model.config.prediction_channel_indices] = next_point
            else:
                new_past = next_point

            # update the past observed mask by copying the future
            if this_future_observed_mask is not None and prediction_channel_indices is not None:
                new_past_observed_mask = this_future_observed_mask.clone()
                new_past_observed_mask[:, :, self.model.config.prediction_channel_indices] = True
            else:
                new_past_observed_mask = torch.ones_like(this_future_observed_mask, dtype=torch.bool)

            this_past = torch.cat([this_past[:, model_prediction_length:], new_past], dim=1)
            this_past_observed_mask = (
                torch.cat([this_past_observed_mask[:, model_prediction_length:], new_past_observed_mask], dim=1)
                if this_future_observed_mask is not None
                else None
            )

        output = predicted_sequence

        loss_val = None

        if future_values is not None:
            loss_val = self.loss(output, future_values)
        if not return_dict:
            return tuple(
                v
                for v in [
                    loss_val,
                    output,
                ]
            )

        return RecursivePredictorOutput(
            loss=loss_val,
            prediction_outputs=output,
            backbone_hidden_state=torch.rand(1, 1),
        )



import json
import logging
import os
import sys

import torch
from fastapi import HTTPException


logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)

# logger.addHandler(logging.StreamHandler(sys.stdout))


# clue sagemaker runtime into some things
os.environ["TSFM_ALLOW_LOAD_FROM_HF_HUB"] = "0"
os.environ["TSFM_MODEL_DIR"] = os.path.join(os.path.dirname(__file__), "granite-tsfm/services/inference/mytest-tsfm")

# service components are not git installable directly
# so hack it a bit this way
custom_path = os.path.join(os.path.dirname(__file__), "granite-tsfm/services/inference")
if custom_path not in sys.path:
    sys.path.append(custom_path)

device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
# currently we're targeting only cpu devices for inference as
# the tsfm models are compact and performant for CPU-only inference
device = "cpu"


# defining model and loading weights to it.
def model_fn(model_dir):
    logger.debug(f"in model_fn with {model_dir}")
    logger.debug("doing nothing, returning {}")
    return {}


# data preprocessing
def input_fn(request_body, request_content_type):
    assert request_content_type == "application/json"
    if logger.isEnabledFor(logging.DEBUG):
        logger.debug("in input_fn")
        logger.debug(f"request_content_type {request_content_type}")
        logger.debug(f"type of request_body {type(request_body)}")
        logger.debug(f"request_body {request_body[0:100]}...")
    return request_body


# inference
def predict_fn(input_object, model):
    from json import JSONDecodeError

    from pydantic import ValidationError
    from tsfminference.inference import InferenceRuntime
    from tsfminference.inference_payloads import (
        ForecastingInferenceInput,
        PredictOutput,
    )

    logger.debug("in predict_fn")
    logger.debug(f"input_object type {type(input_object)}")
    logger.debug(f"model {model}")

    try:
        input: dict = json.loads(input_object)
        logger.debug(f"input is now of type f{type(input)}")
        inference_type = input.pop("inference_type")
        if not "forecasting" == inference_type:
            return (
                json.dumps(
                    {
                        "error_code": "ValidationError",
                        "error_message": f"inference_type {inference_type} is not supported.",
                    }
                ),
                "application/json",
                400,
            )

        input: ForecastingInferenceInput = ForecastingInferenceInput(**input)

        runtime: InferenceRuntime = InferenceRuntime()
        answer: PredictOutput = runtime.forecast(input=input)
        return answer
    except HTTPException as httpex:
        error_response = {"error_code": "HTTPException", "error_message": str(httpex)}
        return json.dumps(error_response), "application/json", 400
    except ValidationError as vex:
        error_response = {"error_code": "ValidationError", "error_message": str(vex)}
        return json.dumps(error_response), "application/json", 400
    except JSONDecodeError as jde:
        error_response = {"error_code": "JSONDecodeError", "error_message": str(jde)}
        return json.dumps(error_response), "application/json", 400
    except Exception as ex:
        error_response = {"error_code": "SERVER_ERROR", "error_message": str(ex)}
        return json.dumps(error_response), "application/json", 500


# postprocess
def output_fn(predictions, content_type):
    assert content_type == "application/json"
    from tsfminference.inference_payloads import (
        PredictOutput,
    )

    if isinstance(predictions, PredictOutput):
        return predictions.model_dump_json()
    else:
        return predictions



import csv
import math
import os
import tempfile
import warnings

import numpy as np
import pandas as pd
import torch
from torch.optim import AdamW
from torch.optim.lr_scheduler import OneCycleLR
from torch.utils.data import DataLoader
from tqdm import tqdm
from transformers import EarlyStoppingCallback, Trainer, TrainingArguments, set_seed
from utils import mask_generate, mse

from tsfm_public import TimeSeriesPreprocessor, get_datasets
from tsfm_public.models.tspulse import TSPulseForReconstruction
from tsfm_public.toolkit.lr_finder import optimal_lr_finder


warnings.filterwarnings("ignore")

device = "cuda"
CONTEXT_LEN = 512
FORECAST_LEN = 0


def main(DATASET, mask_type, mask_ratio):
    seed = 42
    set_seed(seed)
    # Dataset
    if DATASET in ["ETTh1", "ETTh2", "ETTm1", "ETTm2"]:
        dataset_path = f"https://raw.githubusercontent.com/zhouhaoyi/ETDataset/main/ETT-small/{DATASET}.csv"
    else:
        dataset_path = f"datasets/{DATASET}/{DATASET}.csv"

    timestamp_column = "date"
    id_columns = []  # mention the ids that uniquely identify a time-series.

    if DATASET in ["ETTh1", "ETTh2"]:
        split_config = {
            "train": [0, 8640],
            "valid": [8640, 11520],
            "test": [
                11520,
                14400,
            ],
        }
    elif DATASET in ["ETTm1", "ETTm2"]:
        split_config = {
            "train": [0, 34560],
            "valid": [34560, 46080],
            "test": [
                46080,
                57600,
            ],
        }
    else:
        split_config = {
            "train": 0.7,
            "test": 0.2,
        }

    data = pd.read_csv(
        dataset_path,
        parse_dates=[timestamp_column],
    )

    target_columns = data.columns.to_list()[1:]  # all the columns from the data except 'date'

    column_specifiers = {
        "timestamp_column": timestamp_column,
        "id_columns": id_columns,
        "target_columns": target_columns,
        "control_columns": [],
    }

    tsp = TimeSeriesPreprocessor(
        **column_specifiers,
        context_length=CONTEXT_LEN,
        prediction_length=FORECAST_LEN,
        scaling=True,
        encode_categorical=False,
        scaler_type="standard",
    )

    train_dataset, valid_dataset, test_dataset = get_datasets(tsp, data, split_config)

    model_dict = {
        "mask_ratio": mask_ratio,
        "mask_type": mask_type,
        "prediction_length": 0,
        "fft_time_add_forecasting_pt_loss": False,
        "enable_fft_prob_loss": False,
        "fft_time_consistent_masking": True,
        "fft_original_signal_loss_weight": 0,
        "loss_apply_mode": "mask",
        "fft_weight": 0,
        "num_full_patches_for_hybrid_mask": int((mask_ratio / 0.125) * 4),
        "decoder_mode": "mix_channel",
        "channel_consistent_masking": False,
        "dropout": 0,
        "head_dropout": 0,
    }

    model_dict["num_input_channels"] = tsp.num_input_channels

    model = TSPulseForReconstruction.from_pretrained(
        "ibm-granite/granite-timeseries-tspulse-r1", revision="tspulse-hybrid-dualhead-512-p8-r1", **model_dict
    ).to(device)

    OUT_DIR = "tspulse_finetuned_models/"

    model = model.to("cuda").float()

    for param in model.parameters():
        param.requires_grad = True

    temp_dir = tempfile.mkdtemp()

    suggested_lr = None

    train_dict = {
        "overwrite_output_dir": True,
        "learning_rate": 0.0001,
        "num_train_epochs": 100,
        "evaluation_strategy": "epoch",
        "per_device_train_batch_size": 8,
        "per_device_eval_batch_size": 8,
        "dataloader_num_workers": 1,
        "eval_accumulation_steps": 50,
        "ddp_find_unused_parameters": False,
        "report_to": "tensorboard",
        "save_strategy": "epoch",
        "logging_strategy": "epoch",
        "save_total_limit": 1,
        "load_best_model_at_end": True,
        "metric_for_best_model": "eval_loss",
        "greater_is_better": False,
        "seed": 42,
    }

    EPOCHS = train_dict["num_train_epochs"]
    BATCH_SIZE = train_dict["per_device_train_batch_size"]
    eval_accumulation_steps = train_dict["eval_accumulation_steps"]
    NUM_WORKERS = 1
    NUM_GPUS = 1

    set_seed(42)
    if suggested_lr is None:
        lr, model = optimal_lr_finder(
            model,
            train_dataset,
            batch_size=BATCH_SIZE,
        )
        suggested_lr = lr

    finetune_args = TrainingArguments(
        output_dir=temp_dir,
        overwrite_output_dir=True,
        learning_rate=suggested_lr,
        num_train_epochs=EPOCHS,
        do_eval=True,
        eval_strategy="epoch",
        per_device_train_batch_size=BATCH_SIZE,
        per_device_eval_batch_size=BATCH_SIZE,
        eval_accumulation_steps=eval_accumulation_steps,
        dataloader_num_workers=NUM_WORKERS,
        report_to="tensorboard",
        save_strategy="epoch",
        logging_strategy="epoch",
        save_total_limit=1,
        logging_dir=os.path.join(OUT_DIR, "output"),  # Make sure to specify a logging directory
        load_best_model_at_end=True,  # Load the best model when training ends
        metric_for_best_model="eval_loss",  # Metric to monitor for early stopping
        greater_is_better=False,  # For loss
    )

    early_stopping_callback = EarlyStoppingCallback(
        early_stopping_patience=10,  # Number of epochs with no improvement after which to stop
        early_stopping_threshold=0.0001,  # Minimum improvement required to consider as improvement
    )

    # Optimizer and scheduler
    optimizer = AdamW(model.parameters(), lr=suggested_lr)
    scheduler = OneCycleLR(
        optimizer,
        suggested_lr,
        epochs=EPOCHS,
        steps_per_epoch=math.ceil(len(train_dataset) / (BATCH_SIZE * NUM_GPUS)),
    )

    finetune_trainer = Trainer(
        model=model,
        args=finetune_args,
        train_dataset=train_dataset,
        eval_dataset=valid_dataset,
        callbacks=[early_stopping_callback],
        optimizers=(optimizer, scheduler),
    )

    # Fine tune
    finetune_trainer.train()

    # save the finetuned model
    os.makedirs("finetuned_models", exist_ok=True)
    path_to_save_model = f"finetuned_models/finetuned_model_{DATASET}_{mask_ratio}_{mask_type}"
    finetune_trainer.save_model(path_to_save_model)

    if DATASET in ["ETTh1", "ETTh2", "ETTm1", "ETTm2"]:
        batch_size = 64
    else:
        batch_size = 4

    def collate_only_past_values(batch):
        return torch.stack([item["past_values"] for item in batch])

    test_dataloader = DataLoader(
        test_dataset, batch_size=batch_size, shuffle=False, collate_fn=collate_only_past_values
    )

    model_path = path_to_save_model

    # load the finetuned model
    model = TSPulseForReconstruction.from_pretrained(
        model_path, fft_time_add_forecasting_pt_loss=False, num_input_channels=tsp.num_input_channels, mask_type="user"
    ).to(device)

    seed = 42
    g = torch.Generator(device=device)
    g.manual_seed(seed)

    trues, preds, masks = [], [], []
    with torch.no_grad():
        for batch in tqdm(test_dataloader):
            batch_x = batch.to(device)  # b l c

            mask = mask_generate(g, batch_x, 8, mask_ratio, mask_type)

            output = model(past_values=batch_x, past_observed_mask=~mask)

            reconstructed_output = output.reconstruction_outputs

            trues.append(batch_x.detach().cpu().numpy())
            preds.append(reconstructed_output.detach().cpu().numpy())
            masks.append(mask.detach().cpu().numpy())

        preds = np.concatenate(preds)
        trues = np.concatenate(trues)
        masks = np.concatenate(masks)

        MSE = mse(y=trues[masks == 1], y_hat=preds[masks == 1], reduction="mean")
        print(f"Dataset = {DATASET}  : Mask Type = {mask_type}  : Mask Ratio = {mask_ratio}")
        print(f"Mean Squarred Error (MSE)={MSE:.3f}")

        output_file = "tspulse_finetuned_imputation_results.csv"

        if not os.path.exists(output_file):
            with open(output_file, mode="w", newline="") as file:
                writer = csv.writer(file)
                writer.writerow(["Dataset", "Mask Type", "Mask Ratio", "MSE"])

        with open(output_file, mode="a", newline="") as file:
            writer = csv.writer(file)
            writer.writerow([DATASET, mask_type, mask_ratio, f"{MSE:.3f}"])


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--datasets",
        nargs="+",
        type=str,
        required=True,
        choices=["ETTh1", "ETTh2", "ETTm1", "ETTm2", "weather", "electricity"],
        help="List of UEA dataset names",
    )
    parser.add_argument(
        "--mask_type",
        nargs="+",
        type=str,
        required=True,
        choices=["block", "hybrid"],
        help="Masking strategy to evaluate. Options available : 'block' and 'hybrid'",
    )
    parser.add_argument(
        "--mask_ratios",
        nargs="+",
        type=float,
        required=True,
        choices=[0.125, 0.25, 0.375, 0.5],
        help="Masking ratios to evaluate. Options available : 0.125, 0.25, 0.375, 0.5",
    )
    args = parser.parse_args()

    for m_t in args.mask_type:
        for DATASET in args.datasets:
            for m_r in args.mask_ratios:
                main(DATASET, m_t, m_r)



"""Utilities used by test cases in this folder"""

import json
import os

import pandas as pd
import wget
from tsfmfinetuning.ioutils import _split_text_file


def fetch_etth():
    if not os.path.exists("data/ETTH1.csv"):
        dataset_path = "https://raw.githubusercontent.com/zhouhaoyi/ETDataset/main/ETT-small/ETTh1.csv"
        os.makedirs("data", exist_ok=True)
        dataset_path = wget.download(dataset_path, "data/ETTh1.csv")
        timestamp_column = "date"
        data = pd.read_csv(
            dataset_path,
            parse_dates=[timestamp_column],
        )
        data.to_feather("data/ETTh1.feather")
        data.to_csv("data/ETTh1.csv.gz", compression="gzip")


def fetch_finetune_json():
    """Doesn't really fetch, generates it from what fetch_etth"""
    payload = {
        "data": "file://./data/ETTh1.csv",
        "model_id": "mytest-tsfm/ttm-r1",
        "schema": {
            "timestamp_column": "date",
            "target_columns": ["HUFL", "HULL", "MUFL", "MULL", "LUFL", "LULL", "OT"],
        },
        "parameters": {
            "tune_prefix": "fine_tuned/",
            "trainer_args": {"num_train_epochs": 1, "per_device_train_batch_size": 256},
            "fewshot_fraction": 0.05,
        },
    }
    json.dump(payload, open("data/ftpayload.json", "w"))


if __name__ == "__main__":
    fetch_etth()
    fetch_finetune_json()
    _split_text_file(source="./data/ETTh1.csv", target_dir="data/multipart", has_header=True, shuffle=False, parts=5)



# Copyright contributors to the TSFM project
#

# Standard
import json
import os
from datetime import datetime

# Third Party
from transformers import TrainerCallback


class FileLoggingCallback(TrainerCallback):
    """Exports metrics, e.g., training loss to a file in the checkpoint directory."""

    def __init__(self, logs_filename):
        self.training_logs_filename = logs_filename

    def on_log(self, args, state, control, logs=None, **kwargs):
        """Checks if this log contains keys of interest, e.g., loss, and if so, creates
        training_logs.jsonl in the model output dir (if it doesn't already exist),
        appends the subdict of the log & dumps the file.
        """
        # All processes get the logs from this node; only update from process 0.
        if not state.is_world_process_zero:
            return

        log_file_path = os.path.join(args.output_dir, self.training_logs_filename)
        if logs is not None and "loss" in logs and "epoch" in logs:
            self._track_loss("loss", "training_loss", log_file_path, logs, state)
        elif logs is not None and "eval_loss" in logs and "epoch" in logs:
            self._track_loss("eval_loss", "validation_loss", log_file_path, logs, state)

    def _track_loss(self, loss_key, log_name, log_file, logs, state):
        try:
            # Take the subdict of the last log line; if any log_keys aren't part of this log
            # object, assume this line is something else, e.g., train completion, and skip.
            log_obj = {
                "name": log_name,
                "data": {
                    "epoch": round(logs["epoch"], 2),
                    "step": state.global_step,
                    "value": logs[loss_key],
                    "timestamp": datetime.isoformat(datetime.now()),
                },
            }
        except KeyError:
            return

        # append the current log to the jsonl file
        with open(log_file, "a", encoding="utf-8") as f:
            f.write(f"{json.dumps(log_obj, sort_keys=True)}\n")



#!/usr/bin/env python
# coding: utf-8

import logging
import math
import os
import tempfile

import pandas as pd
from torch.optim import AdamW
from torch.optim.lr_scheduler import OneCycleLR
from transformers import EarlyStoppingCallback, Trainer, TrainingArguments, set_seed

from tsfm_public import TimeSeriesPreprocessor, get_datasets
from tsfm_public.models.tinytimemixer import (
    TinyTimeMixerConfig,
    TinyTimeMixerForPrediction,
)
from tsfm_public.models.tinytimemixer.utils import get_ttm_args
from tsfm_public.toolkit.get_model import get_model
from tsfm_public.toolkit.lr_finder import optimal_lr_finder
from tsfm_public.toolkit.visualization import plot_predictions


logger = logging.getLogger(__file__)

# TTM pre-training example.
# This scrips provides a toy example to pretrain a Tiny Time Mixer (TTM) model on
# the `etth1` dataset. For pre-training TTM on a much large set of datasets, please
# have a look at our paper: https://arxiv.org/pdf/2401.03955.pdf
# If you want to directly utilize the pre-trained models. Please use them from the
# Hugging Face Hub: https://huggingface.co/ibm-granite/granite-timeseries-ttm-v1
# Have a look at the fine-tune scripts for example usecases of the pre-trained
# TTM models.

# Basic usage:
# python ttm_pretrain_sample.py --data_root_path datasets/
# See the get_ttm_args() function to know more about other TTM arguments


def get_base_model(args):
    # Pre-train a `TTM` forecasting model
    config = TinyTimeMixerConfig(
        context_length=args.context_length,
        prediction_length=args.forecast_length,
        patch_length=args.patch_length,
        num_input_channels=1,
        patch_stride=args.patch_length,
        d_model=args.d_model,
        num_layers=args.num_layers,  # increase the number of layers if we want more complex models
        mode="common_channel",
        expansion_factor=2,
        dropout=args.dropout,
        head_dropout=args.head_dropout,
        scaling="std",
        gated_attn=True,
        adaptive_patching_levels=args.adaptive_patching_levels,
        # decoder params
        decoder_num_layers=args.decoder_num_layers,  # increase the number of layers if we want more complex models
        decoder_adaptive_patching_levels=0,
        decoder_mode="common_channel",
        decoder_raw_residual=False,
        use_decoder=True,
        decoder_d_model=args.decoder_d_model,
    )

    model = TinyTimeMixerForPrediction(config)
    return model


def pretrain(args, model, dset_train, dset_val):
    # Find optimal learning rate
    # Use with caution: Set it manually if the suggested learning rate is not suitable

    learning_rate, model = optimal_lr_finder(
        model,
        dset_train,
        batch_size=args.batch_size,
    )
    print("OPTIMAL SUGGESTED LEARNING RATE =", learning_rate)

    # learning_rate = args.learning_rate

    trainer_args = TrainingArguments(
        output_dir=os.path.join(args.save_dir, "checkpoint"),
        overwrite_output_dir=True,
        learning_rate=args.learning_rate,
        num_train_epochs=args.num_epochs,
        seed=args.random_seed,
        eval_strategy="epoch",
        per_device_train_batch_size=args.batch_size,
        per_device_eval_batch_size=args.batch_size,
        dataloader_num_workers=args.num_workers,
        ddp_find_unused_parameters=False,
        report_to="tensorboard",
        save_strategy="epoch",
        logging_strategy="epoch",
        save_total_limit=1,
        logging_dir=os.path.join(args.save_dir, "logs"),  # Make sure to specify a logging directory
        load_best_model_at_end=True,  # Load the best model when training ends
        metric_for_best_model="eval_loss",  # Metric to monitor for early stopping
        greater_is_better=False,  # For loss
    )

    # Optimizer and scheduler
    optimizer = AdamW(model.parameters(), lr=learning_rate)
    scheduler = OneCycleLR(
        optimizer,
        learning_rate,
        epochs=args.num_epochs,
        steps_per_epoch=math.ceil(len(dset_train) / args.batch_size),
        # steps_per_epoch=math.ceil(len(dset_train) / (args.batch_size * args.num_gpus)),
    )

    # Create the early stopping callback
    early_stopping_callback = EarlyStoppingCallback(
        early_stopping_patience=10,  # Number of epochs with no improvement after which to stop
        early_stopping_threshold=0.0,  # Minimum improvement required to consider as improvement
    )

    # Set trainer
    if args.early_stopping:
        trainer = Trainer(
            model=model,
            args=trainer_args,
            train_dataset=dset_train,
            eval_dataset=dset_val,
            optimizers=(optimizer, scheduler),
            callbacks=[early_stopping_callback],
        )
    else:
        trainer = Trainer(
            model=model,
            args=trainer_args,
            train_dataset=dset_train,
            eval_dataset=dset_val,
            optimizers=(optimizer, scheduler),
        )

    # Train
    trainer.train()

    # Save the pretrained model

    model_save_path = os.path.join(args.save_dir, "ttm_pretrained")
    trainer.save_model(model_save_path)
    return model_save_path


def inference(args, model_path, dset_test):
    model = get_model(model_path=model_path)

    temp_dir = tempfile.mkdtemp()
    trainer = Trainer(
        model=model,
        args=TrainingArguments(
            output_dir=temp_dir,
            per_device_eval_batch_size=args.batch_size,
            seed=args.random_seed,
            report_to="none",
        ),
    )
    # evaluate = zero-shot performance
    print("+" * 20, "Test MSE output:", "+" * 20)
    output = trainer.evaluate(dset_test)
    print(output)

    # get predictions

    predictions_dict = trainer.predict(dset_test)

    predictions_np = predictions_dict.predictions[0]

    print(predictions_np.shape)

    # get backbone embeddings (if needed for further analysis)

    backbone_embedding = predictions_dict.predictions[1]

    print(backbone_embedding.shape)

    plot_path = os.path.join(args.save_dir, "plots")
    # plot
    plot_predictions(
        model=trainer.model,
        dset=dset_test,
        plot_dir=plot_path,
        plot_prefix="test_inference",
        channel=0,
    )
    print("Plots saved in location:", plot_path)


if __name__ == "__main__":
    # Arguments
    args = get_ttm_args()

    # Set seed
    set_seed(args.random_seed)

    logger.info(
        f"{'*' * 20} Pre-training a TTM for context len = {args.context_length}, forecast len = {args.forecast_length} {'*' * 20}"
    )

    # Data prep
    # Dataset
    TARGET_DATASET = "etth1"
    dataset_path = (
        "https://raw.githubusercontent.com/zhouhaoyi/ETDataset/main/ETT-small/ETTh1.csv"  # mention the dataset path
    )
    timestamp_column = "date"
    id_columns = []  # mention the ids that uniquely identify a time-series.

    target_columns = ["HUFL", "HULL", "MUFL", "MULL", "LUFL", "LULL", "OT"]

    # mention the train, valid and split config.
    split_config = {
        "train": [0, 8640],
        "valid": [8640, 11520],
        "test": [
            11520,
            14400,
        ],
    }

    data = pd.read_csv(
        dataset_path,
        parse_dates=[timestamp_column],
    )

    column_specifiers = {
        "timestamp_column": timestamp_column,
        "id_columns": id_columns,
        "target_columns": target_columns,
        "control_columns": [],
    }

    tsp = TimeSeriesPreprocessor(
        **column_specifiers,
        context_length=args.context_length,
        prediction_length=args.forecast_length,
        scaling=True,
        encode_categorical=False,
        scaler_type="standard",
    )

    dset_train, dset_valid, dset_test = get_datasets(tsp, data, split_config)

    # Get model
    model = get_base_model(args)

    # Pretrain
    model_save_path = pretrain(args, model, dset_train, dset_valid)
    print("=" * 20, "Pretraining Completed!", "=" * 20)
    print("Model saved in location:", model_save_path)

    # inference

    inference(args=args, model_path=model_save_path, dset_test=dset_test)

    print("inference completed..")



# Copyright contributors to the TSFM project
#

"""Tests the time series preprocessor and functions"""

import tempfile
from datetime import datetime, timedelta

import numpy as np
import pandas as pd
import pytest

from tsfm_public.toolkit.time_series_preprocessor import (
    DEFAULT_FREQUENCY_MAPPING,
    OrdinalEncoder,
    StandardScaler,
    TimeSeriesPreprocessor,
    create_timestamps,
    extend_time_series,
    get_datasets,
)
from tsfm_public.toolkit.util import FractionLocation

from ..util import nreps


def test_standard_scaler(sample_data):
    scaler = StandardScaler()

    columns = ["val", "val2"]

    # check shape preserved
    result = scaler.fit_transform(sample_data[columns])
    assert result.shape == sample_data[columns].shape
    expected = (sample_data[columns].values - np.mean(sample_data[columns].values, axis=0)) / np.std(
        sample_data[columns].values, axis=0
    )
    np.testing.assert_allclose(result, expected)

    # check serialization
    state = scaler.to_dict()

    assert StandardScaler.from_dict(state).to_dict() == state


def test_ordinal_encoder(sample_data):
    encoder = OrdinalEncoder()

    columns = ["cat", "cat2"]

    # check shape preserved
    result = encoder.fit_transform(sample_data[columns])
    assert result.shape == sample_data[columns].shape

    expected = np.array(
        [
            [0.0, 0.0],
            [1.0, 1.0],
            [0.0, 2.0],
            [1.0, 3.0],
            [0.0, 4.0],
            [1.0, 0.0],
            [0.0, 1.0],
            [1.0, 2.0],
            [0.0, 3.0],
            [1.0, 4.0],
        ]
    )
    np.testing.assert_allclose(result, expected)

    # check serialization
    state = encoder.to_dict()

    assert OrdinalEncoder.from_dict(state).to_dict() == state


def test_time_series_preprocessor_encodes(sample_data):
    static_categorical_columns = ["cat", "cat2"]

    tsp = TimeSeriesPreprocessor(
        target_columns=["val", "val2"],
        static_categorical_columns=static_categorical_columns,
    )
    tsp.train(sample_data)

    sample_prep = tsp.preprocess(sample_data)

    for c in static_categorical_columns:
        assert sample_prep[c].dtype == float

    # two static categoricals were defined
    assert tsp.categorical_vocab_size_list == [2, 5]

    categorical_columns = ["cat", "cat2"]
    tsp = TimeSeriesPreprocessor(
        target_columns=["val", "val2"],
        categorical_columns=categorical_columns,
        control_columns=categorical_columns,
    )
    tsp.train(sample_data)

    sample_prep = tsp.preprocess(sample_data)

    for c in categorical_columns:
        assert sample_prep[c].dtype == float

    # no static categorical columns defined
    assert tsp.categorical_vocab_size_list is None


def test_time_series_preprocessor_scales(ts_data):
    df = ts_data

    tsp = TimeSeriesPreprocessor(
        timestamp_column="timestamp",
        prediction_length=2,
        context_length=5,
        id_columns=["id", "id2"],
        target_columns=["value1", "value2"],
        scaling=True,
    )

    tsp.train(df)

    # check scaled result
    out = tsp.preprocess(df)
    assert np.allclose(out.groupby(tsp.id_columns)[tsp.target_columns].apply(lambda x: np.mean(x)), 0.0)
    assert np.allclose(out.groupby(tsp.id_columns)[tsp.target_columns].apply(lambda x: np.std(x, axis=0)), 1.0)

    # check inverse scale result
    out_inv = tsp.inverse_scale_targets(out)
    assert np.all(
        out_inv.groupby(tsp.id_columns)[tsp.target_columns].apply(lambda x: np.mean(x))
        == df.groupby(tsp.id_columns)[tsp.target_columns].apply(lambda x: np.mean(x))
    )
    assert np.all(
        out_inv.groupby(tsp.id_columns)[tsp.target_columns].apply(lambda x: np.std(x, axis=0))
        == df.groupby(tsp.id_columns)[tsp.target_columns].apply(lambda x: np.std(x, axis=0))
    )

    # check inverse scale result, with suffix

    suffix = "_foo"
    targets_suffix = [f"{c}{suffix}" for c in tsp.target_columns]
    out.columns = [f"{c}{suffix}" if c in tsp.target_columns else c for c in out.columns]
    out_inv = tsp.inverse_scale_targets(out, suffix=suffix)
    assert np.all(
        out_inv.groupby(tsp.id_columns)[targets_suffix].apply(lambda x: np.mean(x))
        == df.groupby(tsp.id_columns)[tsp.target_columns].apply(lambda x: np.mean(x))
    )


def test_time_series_preprocessor_inv_scales_lists(ts_data):
    df = ts_data

    tsp = TimeSeriesPreprocessor(
        timestamp_column="timestamp",
        prediction_length=2,
        context_length=5,
        id_columns=["id", "id2"],
        target_columns=["value1", "value2"],
        scaling=True,
    )

    tsp.train(df)

    # check scaled result
    out = tsp.preprocess(df)

    # construct artificial result
    out["value1"] = out["value1"].apply(lambda x: np.array([x] * 3))
    out["value2"] = out["value2"].apply(lambda x: np.array([x] * 3))

    out_inv = tsp.inverse_scale_targets(out)

    assert out_inv["value1"].mean()[0] == df["value1"].mean()
    assert out_inv["value2"].mean()[0] == df["value2"].mean()


def test_extend_time_series(ts_data):
    periods = 5
    a = extend_time_series(ts_data, timestamp_column="timestamp", grouping_columns=["id"], periods=periods)

    # check that length increases by periods for each id
    assert a.shape[0] == ts_data.shape[0] + 3 * periods
    assert a.shape[1] == ts_data.shape[1]

    periods = 3
    a = extend_time_series(
        ts_data,
        timestamp_column="timestamp",
        grouping_columns=["id", "id2"],
        periods=periods,
    )

    # check that length increases by periods for each id
    assert a.shape[0] == ts_data.shape[0] + 3 * periods
    assert a.shape[1] == ts_data.shape[1]

    # test different lengths

    ts_data_2 = pd.DataFrame(
        {
            "id": list(nreps(["A", "B"], 50)) + ["C"] * 20,
            "timestamp": [datetime(2021, 1, 1) + timedelta(days=i) for i in range(50)] * 2
            + [datetime(2021, 1, 1) + timedelta(days=i) for i in range(20)],
            "value1": range(120),
        }
    )

    a = extend_time_series(ts_data_2, timestamp_column="timestamp", grouping_columns=["id"], total_periods=60)

    assert len(a) == 180


def test_create_timestamps():
    base_last_timestamp = datetime(2020, 1, 1)
    base_timedelta = timedelta(days=1)
    base_timedelta_strs = ["1d", "1 days 00:00:00", "24h"]
    num_periods = 4
    base_expected = [base_last_timestamp + base_timedelta * i for i in range(1, num_periods + 1)]

    date_types = [datetime, pd.Timestamp, np.datetime64]
    timedelta_types = [timedelta, pd.Timedelta, np.timedelta64]

    test_cases = []
    for date_type in date_types:
        for timedelta_type in timedelta_types:
            test_cases.append(
                {
                    "last_timestamp": base_last_timestamp
                    if isinstance(base_last_timestamp, date_type)
                    else date_type(base_last_timestamp),
                    "freq": base_timedelta
                    if isinstance(base_timedelta, timedelta_type)
                    else timedelta_type(base_timedelta),
                    "periods": 4,
                    "expected": [d if isinstance(d, date_type) else date_type(d) for d in base_expected],
                }
            )

    for date_type in date_types:
        for freq in base_timedelta_strs:
            test_cases.append(
                {
                    "last_timestamp": base_last_timestamp
                    if isinstance(base_last_timestamp, date_type)
                    else date_type(base_last_timestamp),
                    "freq": freq,
                    "periods": 4,
                    "expected": [d if isinstance(d, date_type) else date_type(d) for d in base_expected],
                }
            )

    test_cases.extend(
        [
            {"last_timestamp": 100, "freq": 3, "periods": 2, "expected": [103, 106]},
            {"last_timestamp": 100, "freq": 3.5, "periods": 2, "expected": [103.5, 107.0]},
            {"last_timestamp": 100, "freq": "3.5", "periods": 2, "expected": [103.5, 107.0]},
            {"last_timestamp": np.float32(100), "freq": "3.5", "periods": 2, "expected": [103.5, 107.0]},
        ]
    )

    for test_record in test_cases:
        expected = test_record.pop("expected")
        out = create_timestamps(**test_record)
        assert out == expected


def test_create_timestamps_with_sequence():
    # start, freq, sequence, periods, expected
    test_cases = [
        (
            datetime(2020, 1, 1),
            "1d",
            [datetime(2019, 1, 2), datetime(2019, 1, 3), datetime(2019, 1, 4)],
            2,
            [datetime(2020, 1, 2), datetime(2020, 1, 3)],
        ),
        (
            datetime(2021, 1, 1),
            timedelta(days=365),
            [datetime(2017, 1, 3), datetime(2018, 1, 3), datetime(2019, 1, 3)],
            2,
            [datetime(2022, 1, 1), datetime(2023, 1, 1)],
        ),
        (
            pd.Timestamp(2020, 1, 1),
            pd.Timedelta(weeks=1),
            [pd.Timestamp(2019, 2, 1), pd.Timestamp(2019, 2, 8)],
            2,
            [datetime(2020, 1, 8), datetime(2020, 1, 15)],
        ),
        (
            100,
            3,
            [3, 6, 9],
            2,
            [103, 106],
        ),
        (
            100,
            3.5,
            [10, 13.5, 17.0],
            2,
            [103.5, 107.0],
        ),
        (
            pd.Timestamp(2021, 12, 31),
            "QE",
            None,
            4,
            [
                pd.Timestamp(2022, 3, 31),
                pd.Timestamp(2022, 6, 30),
                pd.Timestamp(2022, 9, 30),
                pd.Timestamp(2022, 12, 31),
            ],
        ),
    ]

    for start, freq, sequence, periods, expected in test_cases:
        # test based on provided freq
        ts = create_timestamps(start, freq=freq, periods=periods)
        assert ts == expected

        # test based on provided sequence
        if sequence is not None:
            ts = create_timestamps(start, time_sequence=sequence, periods=periods)
            assert ts == expected

    # it is an error to provide neither freq or sequence
    with pytest.raises(ValueError):
        ts = create_timestamps(start, periods=periods)


def test_get_datasets(ts_data):
    tsp = TimeSeriesPreprocessor(
        id_columns=["id"],
        target_columns=["value1", "value2"],
        prediction_length=5,
        context_length=10,
    )

    train, valid, test = get_datasets(
        tsp,
        ts_data,
        split_config={"train": [0, 1 / 3], "valid": [1 / 3, 2 / 3], "test": [2 / 3, 1]},
    )

    # 3 time series of length 50
    assert len(train) == 3 * (int((1 / 3) * 50) - (tsp.context_length + tsp.prediction_length) + 1)

    assert len(valid) == len(test)

    full_lengths = [len(train), len(valid), len(test)]

    stride = 3
    num_ids = len(ts_data["id"].unique())
    # test stride
    train, valid, test = get_datasets(
        tsp, ts_data, split_config={"train": [0, 1 / 3], "valid": [1 / 3, 2 / 3], "test": [2 / 3, 1]}, stride=stride
    )

    strided_lengths = [len(train), len(valid), len(test)]

    # x is full length under stride 1
    # x // 3 is full length for each ID, need to subtract one and then compute strided length per ID
    assert [(((x // num_ids) - 1) // stride + 1) * num_ids for x in full_lengths] == strided_lengths

    # no id columns, so treat as one big time series
    tsp = TimeSeriesPreprocessor(
        id_columns=[],
        target_columns=["value1", "value2"],
        prediction_length=5,
        context_length=10,
    )

    train, valid, test = get_datasets(
        tsp,
        ts_data,
        split_config={
            "train": [0, 100],
            "valid": [100, 125],
            "test": [125, 150],
        },
        fewshot_fraction=0.2,
        fewshot_location=FractionLocation.LAST.value,
    )

    # new train length should be 20% of 100, minus the usual for context length and prediction length
    fewshot_train_size = (
        int((100 - tsp.context_length) * 0.2) + tsp.context_length - (tsp.context_length + tsp.prediction_length) + 1
    )
    assert len(train) == fewshot_train_size

    assert len(valid) == len(test)

    # no id columns, so treat as one big time series
    tsp = TimeSeriesPreprocessor(
        id_columns=[],
        target_columns=["value1", "value2"],
        prediction_length=5,
        context_length=10,
    )

    train, valid, test = get_datasets(
        tsp,
        ts_data,
        split_config={
            "train": [0, 100],
            "valid": [100, 125],
            "test": [125, 150],
        },
        fewshot_fraction=0.2,
        fewshot_location=FractionLocation.FIRST.value,
    )

    # new train length should be 20% of 100, minus the usual for context length and prediction length
    assert len(train) == fewshot_train_size

    assert len(valid) == len(test)

    # fraction splits
    # no id columns, so treat as one big time series
    tsp = TimeSeriesPreprocessor(
        id_columns=[],
        target_columns=["value1", "value2"],
        prediction_length=5,
        context_length=10,
    )

    train, valid, test = get_datasets(
        tsp,
        ts_data,
        split_config={
            "train": 0.7,
            "test": 0.2,
        },
    )

    assert len(train) == int(150 * 0.7) - (tsp.context_length + tsp.prediction_length) + 1

    assert len(test) == int(150 * 0.2) - tsp.prediction_length + 1

    assert len(valid) == 150 - int(150 * 0.2) - int(150 * 0.7) - tsp.prediction_length + 1

    full_train_size = len(train)

    train, valid, test = get_datasets(
        tsp,
        ts_data,
        split_config={
            "train": 0.7,
            "test": 0.2,
        },
        fewshot_fraction=0.2,
        fewshot_location=FractionLocation.UNIFORM.value,
        seed=42,
    )

    assert len(train) == int(full_train_size * 0.2)


def test__validate_columns(ts_data):
    # should work
    TimeSeriesPreprocessor(
        id_columns=["id"],
        target_columns=["value1", "value2"],
        prediction_length=5,
        context_length=13,
    ).train(ts_data)

    # should not work
    with pytest.raises(ValueError):
        TimeSeriesPreprocessor(
            id_columns=["id"],
            target_columns=["value1", "value2"],
            control_columns=["value1"],
            prediction_length=5,
            context_length=13,
        ).train(ts_data)


def test_get_datasets_padding(ts_data):
    tsp = TimeSeriesPreprocessor(
        id_columns=["id"],
        target_columns=["value1", "value2"],
        prediction_length=5,
        context_length=13,
    )

    train, valid, test = get_datasets(
        tsp,
        ts_data,
        split_config={"train": [0, 1 / 3], "valid": [1 / 3, 2 / 3], "test": [2 / 3, 1]},
    )

    assert len(train) == 3

    with pytest.raises(RuntimeError):
        train, valid, test = get_datasets(
            tsp,
            ts_data,
            split_config={"train": [0, 1 / 3], "valid": [1 / 3, 2 / 3], "test": [2 / 3, 1]},
            enable_padding=False,
        )


def test_train_without_targets(ts_data):
    # no targets or other columns specified
    tsp = TimeSeriesPreprocessor(id_columns=["id", "id2"], timestamp_column="timestamp")
    tsp.train(ts_data)

    assert tsp.target_columns == ["value1", "value2"]

    # some other args specified
    for arg in [
        "control_columns",
        "conditional_columns",
        "observable_columns",
        "static_categorical_columns",
    ]:
        tsp = TimeSeriesPreprocessor(
            id_columns=["id", "id2"],
            timestamp_column="timestamp",
            **{arg: ["value2"]},
        )
        tsp.train(ts_data)

        assert tsp.target_columns == ["value1"]

    # test targets honored
    tsp = TimeSeriesPreprocessor(
        id_columns=["id", "id2"],
        timestamp_column="timestamp",
        target_columns=["value2"],
    )
    tsp.train(ts_data)

    assert tsp.target_columns == ["value2"]


def test_get_datasets_without_targets(ts_data):
    ts_data = ts_data.drop(columns=["id", "id2"])
    tsp = TimeSeriesPreprocessor(
        timestamp_column="timestamp",
        prediction_length=2,
        context_length=5,
    )

    train, _, _ = get_datasets(tsp, ts_data, split_config={"train": 0.7, "test": 0.2})

    assert train.datasets[0].target_columns == ["value1", "value2"]


def test_get_datasets_univariate(ts_data):
    tsp = TimeSeriesPreprocessor(
        timestamp_column="timestamp",
        id_columns=["id", "id2"],
        target_columns=["value1", "value2"],
        prediction_length=2,
        context_length=5,
    )

    # for baseline
    train_base, _, _ = get_datasets(tsp, ts_data, split_config={"train": 0.7, "test": 0.2})

    train, _, _ = get_datasets(tsp, ts_data, split_config={"train": 0.7, "test": 0.2}, as_univariate=True)

    assert train[0]["id"][-1] == "value1"
    assert len(train) == 2 * len(train_base)


def test_get_datasets_with_frequency_token(ts_data):
    ts_data = ts_data.drop(columns=["id", "id2"])
    tsp = TimeSeriesPreprocessor(timestamp_column="timestamp", prediction_length=2, context_length=5, freq="d")

    train, _, _ = get_datasets(tsp, ts_data, split_config={"train": 0.7, "test": 0.2}, use_frequency_token=True)

    assert train[0]["freq_token"] == DEFAULT_FREQUENCY_MAPPING["d"]


def test_masking_specification(ts_data):
    ts_data = ts_data.drop(columns=["id", "id2"])
    tsp = TimeSeriesPreprocessor(timestamp_column="timestamp", prediction_length=2, context_length=5, freq="d")

    train, _, _ = get_datasets(
        tsp,
        ts_data,
        split_config={"train": 0.7, "test": 0.2},
        use_frequency_token=True,
        fill_value=-1000,
        masking_specification=[("value1", -2)],
    )

    assert np.all(train[0]["past_values"].numpy()[-2:, 0] == -1000)


def test_get_frequency_token():
    tsp = TimeSeriesPreprocessor(timestamp_column="date")

    assert tsp.get_frequency_token("1h") == DEFAULT_FREQUENCY_MAPPING["h"]
    assert tsp.get_frequency_token("h") == DEFAULT_FREQUENCY_MAPPING["h"]
    assert tsp.get_frequency_token("0 days 01:00:00") == DEFAULT_FREQUENCY_MAPPING["h"]
    assert tsp.get_frequency_token("H") == DEFAULT_FREQUENCY_MAPPING["h"]
    assert tsp.get_frequency_token("1H") == DEFAULT_FREQUENCY_MAPPING["h"]


def test_id_columns_and_scaling_id_columns(ts_data_runs):
    df = ts_data_runs

    tsp = TimeSeriesPreprocessor(
        timestamp_column="timestamp",
        prediction_length=2,
        context_length=5,
        id_columns=["asset_id", "run_id"],
        scaling_id_columns=["asset_id"],
        target_columns=["value1"],
        scaling=True,
    )

    ds_train, _, _ = get_datasets(tsp, df, split_config={"train": 0.7, "test": 0.2})

    assert len(tsp.target_scaler_dict) == 2
    assert len(ds_train.datasets) == 4

    tsp = TimeSeriesPreprocessor(
        timestamp_column="timestamp",
        prediction_length=2,
        context_length=5,
        id_columns=["asset_id", "run_id"],
        scaling_id_columns=[],
        target_columns=["value1"],
        scaling=True,
    )

    ds_train, _, _ = get_datasets(tsp, df, split_config={"train": 0.7, "test": 0.2})

    assert len(tsp.target_scaler_dict) == 1
    assert len(ds_train.datasets) == 4

    tsp = TimeSeriesPreprocessor(
        timestamp_column="timestamp",
        prediction_length=2,
        context_length=5,
        id_columns=["asset_id", "run_id"],
        scaling_id_columns=None,
        target_columns=["value1"],
        scaling=True,
    )

    ds_train, _, _ = get_datasets(tsp, df, split_config={"train": 0.7, "test": 0.2})

    assert len(tsp.target_scaler_dict) == 4
    assert len(ds_train.datasets) == 4


def test_get_datasets_with_categoricical(ts_data):
    ts_data = ts_data.copy()

    ts_data["varying_cat"] = ts_data.apply(lambda x: x["id"] + str(int(x["value1"] % 5)), axis=1)

    tsp = TimeSeriesPreprocessor(
        timestamp_column="timestamp",
        id_columns=["id", "id2"],
        target_columns=["value1", "value2"],
        prediction_length=2,
        context_length=5,
        categorical_columns=["varying_cat"],
        conditional_columns=["varying_cat"],
        scaling=False,
    )

    # for baseline
    train, _, _ = get_datasets(tsp, ts_data, split_config={"train": 0.7, "test": 0.2})
    expected = np.array([2.0, 3.0, 4.0, 0.0, 1.0])
    np.testing.assert_allclose(train[2]["past_values"][:, 2].numpy(), expected)

    tsp = TimeSeriesPreprocessor(
        timestamp_column="timestamp",
        id_columns=["id", "id2"],
        target_columns=["value1", "value2"],
        prediction_length=2,
        context_length=5,
        categorical_columns=["varying_cat"],
        conditional_columns=["varying_cat"],
        scaling=True,
    )

    # for baseline
    train, _, _ = get_datasets(tsp, ts_data, split_config={"train": 0.7, "test": 0.2})
    expected = np.array([0.0000, 0.7071, 1.4142, -1.4142, -0.7071])
    np.testing.assert_allclose(train[2]["past_values"][:, 2].numpy(), expected, rtol=1e-4)


def test_time_series_preprocessor_serializes(ts_data_runs):
    # two string ids
    df = ts_data_runs.copy()

    tsp = TimeSeriesPreprocessor(
        timestamp_column="timestamp",
        prediction_length=2,
        context_length=5,
        id_columns=["asset_id", "run_id"],
        target_columns=["value1"],
        scaling=True,
    )
    tsp.train(df)

    with tempfile.TemporaryDirectory() as d:
        tsp.save_pretrained(d)
        new_tsp = TimeSeriesPreprocessor.from_pretrained(d)
        assert new_tsp.target_scaler_dict.keys() == tsp.target_scaler_dict.keys()

    # mixed, str int
    df = ts_data_runs.copy()
    df["run_id2"] = df["run_id"].astype(int)

    tsp = TimeSeriesPreprocessor(
        timestamp_column="timestamp",
        prediction_length=2,
        context_length=5,
        id_columns=["asset_id", "run_id2"],
        target_columns=["value1"],
        scaling=True,
    )
    tsp.train(df)

    with tempfile.TemporaryDirectory() as d:
        tsp.save_pretrained(d)
        new_tsp = TimeSeriesPreprocessor.from_pretrained(d)
        assert new_tsp.target_scaler_dict.keys() == tsp.target_scaler_dict.keys()

    # mixed, str int float
    df = ts_data_runs.copy()
    df["run_id2"] = df["run_id"].astype(int)
    df["run_id3"] = df["run_id"].astype(float)

    tsp = TimeSeriesPreprocessor(
        timestamp_column="timestamp",
        prediction_length=2,
        context_length=5,
        id_columns=["asset_id", "run_id2", "run_id3"],
        target_columns=["value1"],
        scaling=True,
    )

    with pytest.raises(Exception):
        tsp.train(df)

    # no scaling
    df = ts_data_runs.copy()

    tsp = TimeSeriesPreprocessor(
        timestamp_column="timestamp",
        prediction_length=2,
        context_length=5,
        target_columns=["value1"],
        id_columns=["asset_id", "run_id"],
        scaling=False,
    )
    tsp.train(df)

    with tempfile.TemporaryDirectory() as d:
        tsp.save_pretrained(d)
        new_tsp = TimeSeriesPreprocessor.from_pretrained(d)
        assert len(new_tsp.target_scaler_dict.keys()) == 0

    # no ids
    df = ts_data_runs.copy()
    tsp = TimeSeriesPreprocessor(
        timestamp_column="timestamp",
        prediction_length=2,
        context_length=5,
        target_columns=["value1"],
        scaling=True,
    )
    tsp.train(df)

    with tempfile.TemporaryDirectory() as d:
        tsp.save_pretrained(d)
        new_tsp = TimeSeriesPreprocessor.from_pretrained(d)
        assert len(new_tsp.target_scaler_dict.keys()) == len(tsp.target_scaler_dict.keys())
        assert len(new_tsp.target_scaler_dict.keys()) == 1



# Copyright contributors to the TSFM project
#
import json
import os
from typing import Any, Dict, Optional, Union

import numpy as np
import pandas as pd
import pytest
import requests

from tsfm_public.toolkit.time_series_preprocessor import extend_time_series
from tsfm_public.toolkit.util import encode_data, select_by_index


DUMPPAYLOADS = int(os.getenv("TSFM_TESTS_DUMP_PAYLOADS", "0")) == 1

model_param_map = {
    "ttm-r1": {"context_length": 512, "prediction_length": 96},
    "ttm-1024-96-r1": {"context_length": 1024, "prediction_length": 96},
    "ttm-r2": {"context_length": 512, "prediction_length": 96},
    "ttm-r2-etth-finetuned": {"context_length": 512, "prediction_length": 96},
    "ttm-r2-etth-finetuned-control": {"context_length": 512, "prediction_length": 96},
    "ttm-1024-96-r2": {"context_length": 1024, "prediction_length": 96},
    "ttm-1536-96-r2": {"context_length": 1536, "prediction_length": 96},
    "ibm/test-patchtst": {"context_length": 512, "prediction_length": 96},
    "ibm/test-patchtsmixer": {"context_length": 512, "prediction_length": 96},
    "ttm-r2-etth-finetuned-impute": {"context_length": 512, "prediction_length": 96},
    "ttm-r2-no-tsfm-config": {"context_length": 512, "prediction_length": 96},
}


@pytest.fixture(scope="module")
def ts_data_base():
    dataset_path = "https://raw.githubusercontent.com/zhouhaoyi/ETDataset/main/ETT-small/ETTh1.csv"

    # forecast_length = 96
    # context_length = 512
    timestamp_column = "date"

    data = pd.read_csv(
        dataset_path,
        parse_dates=[timestamp_column],
    )

    return data


@pytest.fixture(scope="module")
def ts_data(ts_data_base, request):
    # forecast_length = 96
    # context_length = 512
    model_id = request.param
    prediction_length = model_param_map[model_id]["prediction_length"]
    context_length = model_param_map[model_id]["context_length"]
    timestamp_column = "date"

    test_data = select_by_index(
        ts_data_base,
        start_index=12 * 30 * 24 + 4 * 30 * 24 - context_length * 5,
        end_index=12 * 30 * 24 + 4 * 30 * 24,
    ).reset_index(drop=True)

    test_data["id"] = np.array(["a", "b", "c", "d", "e"]).repeat(context_length)

    return test_data, {
        "timestamp_column": timestamp_column,
        "id_columns": ["id"],
        "target_columns": ["HUFL", "HULL", "MUFL", "MULL", "LUFL", "LULL", "OT"],
        "prediction_length": prediction_length,
        "context_length": context_length,
        "model_id": model_id,
    }


def get_inference_response(msg: Dict[str, Any], dumpfile: Optional[Union[str, None]] = None) -> pd.DataFrame:
    URL = (
        "http://127.0.0.1:8000/v1/inference/forecasting"
        if os.environ.get("TSFM_FORECASTING_ENDPOINT", None) is None
        else os.environ.get("TSFM_FORECASTING_ENDPOINT")
    )
    headers = {}
    if dumpfile:
        json.dump(msg, fp=open(dumpfile, "w"), indent=4)
    req = requests.post(URL, json=msg, headers=headers)

    #
    if req.ok:
        resp = req.json()

        df = [pd.DataFrame.from_dict(r) for r in resp["results"]]
        return df, {k: v for k, v in resp.items() if "data_point" in k}
    else:
        print(req.text)
        return req


@pytest.mark.parametrize(
    "ts_data",
    ["ttm-r2-no-tsfm-config"],
    indirect=True,
)
def test_forecast_inference_no_config(ts_data):
    test_data, params = ts_data

    prediction_length = params["prediction_length"]
    context_length = params["context_length"]
    model_id = params["model_id"]
    model_id_path: str = model_id

    id_columns = params["id_columns"]

    # test single
    test_data_ = test_data[test_data[id_columns[0]] == "a"].copy()

    msg = {
        "model_id": model_id_path,
        "parameters": {
            # "prediction_length": params["prediction_length"],
        },
        "schema": {
            "timestamp_column": params["timestamp_column"],
            "id_columns": params["id_columns"],
            "target_columns": params["target_columns"],
        },
        "data": encode_data(test_data_, params["timestamp_column"]),
        "future_data": {},
    }

    df_out, counts = get_inference_response(msg)
    assert len(df_out) == 1
    assert df_out[0].shape[0] == prediction_length
    assert counts["input_data_points"] == context_length * len(params["target_columns"])
    assert counts["output_data_points"] == prediction_length * len(params["target_columns"])


@pytest.mark.parametrize(
    "ts_data",
    ["ttm-r1", "ttm-1024-96-r1", "ttm-r2", "ttm-1024-96-r2", "ttm-1536-96-r2"],
    indirect=True,
)
def test_zero_shot_forecast_inference(ts_data):
    test_data, params = ts_data

    prediction_length = params["prediction_length"]
    context_length = params["context_length"]
    model_id = params["model_id"]
    model_id_path: str = model_id

    id_columns = params["id_columns"]

    num_ids = test_data[id_columns[0]].nunique()

    # test single
    test_data_ = test_data[test_data[id_columns[0]] == "a"].copy()

    msg = {
        "model_id": model_id_path,
        "parameters": {
            # "prediction_length": params["prediction_length"],
        },
        "schema": {
            "timestamp_column": params["timestamp_column"],
            "id_columns": params["id_columns"],
            "target_columns": params["target_columns"],
        },
        "data": encode_data(test_data_, params["timestamp_column"]),
        "future_data": {},
    }

    df_out, counts = get_inference_response(msg, dumpfile="/tmp/test_zero_shot_forecast_inference.json")
    assert len(df_out) == 1
    assert df_out[0].shape[0] == prediction_length
    assert counts["input_data_points"] == context_length * len(params["target_columns"])
    assert counts["output_data_points"] == prediction_length * len(params["target_columns"])

    # test single, very short (length 2)
    test_data_ = test_data[test_data[id_columns[0]] == "a"].copy()

    msg = {
        "model_id": model_id_path,
        "parameters": {
            "prediction_length": params["prediction_length"],
        },
        "schema": {
            "timestamp_column": params["timestamp_column"],
            "id_columns": params["id_columns"],
            "target_columns": params["target_columns"],
        },
        "data": encode_data(test_data_.iloc[:2], params["timestamp_column"]),
        "future_data": {},
    }

    out = get_inference_response(msg)
    assert "Received 2 time points for id a" in out.text

    # test single, more data
    test_data_ = test_data[test_data[id_columns[0]] == "a"].copy()

    test_data_ = extend_time_series(
        test_data_, params["timestamp_column"], grouping_columns=id_columns, freq="1h", periods=10
    )
    test_data_ = test_data_.fillna(0)

    msg = {
        "model_id": model_id_path,
        "parameters": {
            # "prediction_length": params["prediction_length"],
        },
        "schema": {
            "timestamp_column": params["timestamp_column"],
            "id_columns": params["id_columns"],
            "target_columns": params["target_columns"],
        },
        "data": encode_data(test_data_, params["timestamp_column"]),
        "future_data": {},
    }

    df_out, counts = get_inference_response(msg)
    assert len(df_out) == 1
    assert df_out[0].shape[0] == prediction_length
    assert counts["input_data_points"] == context_length * len(params["target_columns"])
    assert counts["output_data_points"] == prediction_length * len(params["target_columns"])

    # test multi-time series
    test_data_ = test_data.copy()

    msg = {
        "model_id": model_id_path,
        "parameters": {
            # "prediction_length": params["prediction_length"],
        },
        "schema": {
            "timestamp_column": params["timestamp_column"],
            "id_columns": params["id_columns"],
            "target_columns": params["target_columns"],
        },
        "data": encode_data(test_data_, params["timestamp_column"]),
        "future_data": {},
    }

    df_out, counts = get_inference_response(msg)

    assert len(df_out) == 1
    assert df_out[0].shape[0] == prediction_length * num_ids
    assert counts["input_data_points"] == context_length * len(params["target_columns"]) * num_ids
    assert counts["output_data_points"] == prediction_length * len(params["target_columns"]) * num_ids

    # test multi-time series, errors
    test_data_ = test_data.copy()
    test_data_ = test_data_.iloc[3:]

    msg = {
        "model_id": model_id_path,
        "parameters": {
            # "prediction_length": params["prediction_length"],
        },
        "schema": {
            "timestamp_column": params["timestamp_column"],
            "id_columns": params["id_columns"],
            "target_columns": params["target_columns"],
        },
        "data": encode_data(test_data_, params["timestamp_column"]),
        "future_data": {},
    }

    out = get_inference_response(msg)
    assert f"Received {context_length - 3} time points for id a" in out.text

    # test multi-time series, multi-id
    # error due to insufficient context
    test_data_ = test_data.copy()
    test_data_ = test_data_.iloc[3:]
    test_data_["id2"] = test_data_[params["id_columns"]]

    msg = {
        "model_id": model_id_path,
        "parameters": {
            # "prediction_length": params["prediction_length"],
        },
        "schema": {
            "timestamp_column": params["timestamp_column"],
            "id_columns": params["id_columns"] + ["id2"],
            "target_columns": params["target_columns"],
        },
        "data": encode_data(test_data_, params["timestamp_column"]),
        "future_data": {},
    }

    out = get_inference_response(msg)
    assert f"Received {context_length - 3} time points for id ('a', 'a')" in out.text

    # single series, less columns
    test_data_ = test_data[test_data[id_columns[0]] == "a"].copy()

    msg = {
        "model_id": model_id_path,
        "parameters": {
            "prediction_length": params["prediction_length"],
        },
        "schema": {
            "timestamp_column": params["timestamp_column"],
            "id_columns": params["id_columns"],
            "target_columns": params["target_columns"][:4],
        },
        "data": encode_data(test_data_, params["timestamp_column"]),
        "future_data": {},
    }

    df_out, counts = get_inference_response(msg)
    assert len(df_out) == 1
    assert df_out[0].shape[0] == prediction_length
    assert df_out[0].shape[1] == 6
    assert counts["input_data_points"] == context_length * len(params["target_columns"][:4])
    assert counts["output_data_points"] == prediction_length * len(params["target_columns"][:4])

    # single series, less columns, no id
    test_data_ = test_data[test_data[id_columns[0]] == "a"].copy()

    msg = {
        "model_id": model_id_path,
        "parameters": {
            "prediction_length": params["prediction_length"],
        },
        "schema": {
            "timestamp_column": params["timestamp_column"],
            "id_columns": [],
            "target_columns": ["HULL"],
        },
        "data": encode_data(test_data_, params["timestamp_column"]),
        "future_data": {},
    }

    df_out, counts = get_inference_response(msg)
    assert len(df_out) == 1
    assert df_out[0].shape[0] == prediction_length
    assert df_out[0].shape[1] == 2
    assert counts["input_data_points"] == context_length
    assert counts["output_data_points"] == prediction_length

    # single series, different prediction length
    test_data_ = test_data[test_data[id_columns[0]] == "a"].copy()

    msg = {
        "model_id": model_id_path,
        "parameters": {
            "prediction_length": params["prediction_length"] // 4,
        },
        "schema": {
            "timestamp_column": params["timestamp_column"],
            "id_columns": params["id_columns"],
            "target_columns": params["target_columns"],
        },
        "data": encode_data(test_data_, params["timestamp_column"]),
        "future_data": {},
    }

    df_out, counts = get_inference_response(msg)
    assert len(df_out) == 1
    assert df_out[0].shape[0] == prediction_length // 4
    assert counts["input_data_points"] == context_length * len(params["target_columns"])
    assert counts["output_data_points"] == (prediction_length // 4) * len(params["target_columns"])

    # single series
    # error wrong prediction length
    test_data_ = test_data[test_data[id_columns[0]] == "a"].copy()

    msg = {
        "model_id": model_id_path,
        "parameters": {
            "prediction_length": params["prediction_length"] * 4,
        },
        "schema": {
            "timestamp_column": params["timestamp_column"],
            "id_columns": params["id_columns"],
            "target_columns": params["target_columns"],
        },
        "data": encode_data(test_data_, params["timestamp_column"]),
        "future_data": {},
    }

    out = get_inference_response(msg)
    assert "prediction_filter_length should be positive" in out.text

    # single series, different prediction length
    test_data_ = test_data[test_data[id_columns[0]] == "a"].copy()

    msg = {
        "model_id": model_id_path,
        "parameters": {
            "prediction_length": params["prediction_length"] // 4,
        },
        "schema": {
            "timestamp_column": params["timestamp_column"],
            "id_columns": params["id_columns"],
            "target_columns": params["target_columns"][1:],
        },
        "data": encode_data(test_data_, params["timestamp_column"]),
        "future_data": {},
    }

    df_out, counts = get_inference_response(msg)
    assert len(df_out) == 1
    assert df_out[0].shape[0] == prediction_length // 4
    assert counts["input_data_points"] == context_length * len(params["target_columns"][1:])
    assert counts["output_data_points"] == (prediction_length // 4) * len(params["target_columns"][1:])


@pytest.mark.parametrize("ts_data", ["ttm-r2-etth-finetuned-control"], indirect=True)
def test_future_data_forecast_inference(ts_data):
    test_data, params = ts_data

    prediction_length = params["prediction_length"]
    context_length = params["context_length"]
    model_id = params["model_id"]
    model_id_path: str = model_id

    id_columns = params["id_columns"]

    num_ids = test_data[id_columns[0]].nunique()

    # test multi series, too short future data
    test_data_ = test_data.copy()

    target_columns = ["OT"]

    future_data = extend_time_series(
        select_by_index(test_data_, id_columns=params["id_columns"], start_index=-1),
        timestamp_column=params["timestamp_column"],
        grouping_columns=params["id_columns"],
        total_periods=25,
        freq="1h",
    )
    future_data = future_data.fillna(0)
    # target data not used for future data (but no harm in keeping it)
    future_data.drop("OT", axis=1, inplace=True)

    prediction_length = 30

    msg = {
        "model_id": model_id_path,
        "parameters": {
            "prediction_length": prediction_length,
        },
        "schema": {
            "timestamp_column": params["timestamp_column"],
            "id_columns": params["id_columns"],
            # "target_columns": target_columns,
            # "control_columns": [c for c in params["target_columns"] if c not in target_columns],
            # "freq": "1h",
        },
        "data": encode_data(test_data_, params["timestamp_column"]),
        "future_data": encode_data(future_data, params["timestamp_column"]),
    }
    out = get_inference_response(
        msg, dumpfile="/tmp/test_future_data_forecast_inference.json" if DUMPPAYLOADS else None
    )
    assert (
        "Future data should have time series of length that is at least the specified prediction length." in out.text
    )

    # test single series, longer future data
    test_data_ = test_data[test_data[id_columns[0]] == "a"].copy()
    num_ids = 1
    # test_data_ = test_data.copy()

    target_columns = ["OT"]

    future_data = extend_time_series(
        select_by_index(test_data_, id_columns=params["id_columns"], start_index=-1),
        timestamp_column=params["timestamp_column"],
        grouping_columns=params["id_columns"],
        total_periods=25,
        freq="1h",
    )
    future_data = future_data.fillna(0)

    prediction_length = 20

    msg = {
        "model_id": model_id_path,
        "parameters": {
            "prediction_length": prediction_length,
        },
        "schema": {
            "timestamp_column": params["timestamp_column"],
            "id_columns": params["id_columns"],
            "target_columns": target_columns,
            "control_columns": [c for c in params["target_columns"] if c not in target_columns],
            "freq": "1h",
        },
        "data": encode_data(test_data_, params["timestamp_column"]),
        "future_data": encode_data(future_data, params["timestamp_column"]),
    }

    df_out, counts = get_inference_response(msg)
    assert len(df_out) == 1
    assert df_out[0].shape[0] == prediction_length * num_ids
    assert (
        counts["input_data_points"]
        == (
            context_length * len(params["target_columns"])
            + prediction_length * (len(params["target_columns"]) - len(target_columns))
        )
        * num_ids
    )
    assert counts["output_data_points"] == prediction_length * 1 * num_ids


@pytest.mark.parametrize("ts_data", ["ttm-r2-etth-finetuned-impute"], indirect=True)
def test_forecast_inference_with_impute(ts_data):
    test_data, params = ts_data

    prediction_length = params["prediction_length"]
    model_id = params["model_id"]
    model_id_path: str = model_id

    id_columns = params["id_columns"]

    # test single series, longer future data
    test_data_ = test_data[test_data[id_columns[0]] == "a"].copy()

    num_ids = 1

    target_columns = ["OT"]
    mask = np.random.rand(len(test_data_)) < 0.1
    test_data_.loc[mask, target_columns[0]] = np.nan

    future_data = extend_time_series(
        select_by_index(test_data_, id_columns=params["id_columns"], start_index=-1),
        timestamp_column=params["timestamp_column"],
        grouping_columns=params["id_columns"],
        total_periods=20,
        freq="1h",
    )
    future_data = future_data.fillna(0)

    prediction_length = 20

    msg = {
        "model_id": model_id_path,
        "parameters": {
            "prediction_length": prediction_length,
        },
        "schema": {
            "timestamp_column": params["timestamp_column"],
            "id_columns": params["id_columns"],
            "target_columns": target_columns,
            "control_columns": [c for c in params["target_columns"] if c not in target_columns],
            "freq": "1h",
        },
        "data": encode_data(test_data_, params["timestamp_column"]),
        "future_data": encode_data(future_data, params["timestamp_column"]),
    }

    df_out, counts = get_inference_response(msg)

    assert len(df_out) == 1
    assert df_out[0].shape[0] == prediction_length * num_ids


@pytest.mark.parametrize(
    "ts_data", ["ttm-r1", "ttm-1024-96-r1", "ttm-r2", "ttm-1024-96-r2", "ttm-1536-96-r2"], indirect=True
)
def test_zero_shot_forecast_inference_no_timestamp(ts_data):
    test_data, params = ts_data

    prediction_length = params["prediction_length"]

    model_id = params["model_id"]
    model_id_path: str = model_id

    id_columns = params["id_columns"]

    # test single with integer as timestamp
    test_data_ = test_data[test_data[id_columns[0]] == "a"].copy()

    test_data_["int_timestamp"] = range(len(test_data_))
    test_data_ = test_data_.drop(columns=params["timestamp_column"])
    msg = {
        "model_id": model_id_path,
        "parameters": {
            # "prediction_length": params["prediction_length"],
        },
        "schema": {
            "timestamp_column": "int_timestamp",
            "id_columns": params["id_columns"],
            "target_columns": params["target_columns"],
        },
        "data": encode_data(test_data_, "int_timestamp"),
        "future_data": {},
    }

    df_out, _ = get_inference_response(msg)
    assert len(df_out) == 1
    assert df_out[0].shape[0] == prediction_length


@pytest.mark.parametrize(
    "ts_data", ["ttm-r1", "ttm-1024-96-r1", "ttm-r2", "ttm-1024-96-r2", "ttm-1536-96-r2"], indirect=True
)
def test_forecast_inference_float_timestamp(ts_data):
    test_data, params = ts_data

    prediction_length = params["prediction_length"]

    model_id = params["model_id"]
    model_id_path: str = model_id

    id_columns = params["id_columns"]

    # test single with float as timestamp
    test_data_ = test_data[test_data[id_columns[0]] == "a"].copy()

    test_data_["float_timestamp"] = [0.0 + x for x in range(len(test_data_))]
    # testing that without being specified if figures it out?
    test_data_ = test_data_.drop(columns=params["timestamp_column"])
    msg = {
        "model_id": model_id_path,
        "parameters": {
            # "prediction_length": params["prediction_length"],
        },
        "schema": {
            "timestamp_column": "float_timestamp",
            "id_columns": params["id_columns"],
            "target_columns": params["target_columns"],
        },
        "data": encode_data(test_data_, "float_timestamp"),
        "future_data": {},
    }

    df_out, _ = get_inference_response(msg)
    assert len(df_out) == 1
    assert df_out[0].shape[0] == prediction_length


@pytest.mark.parametrize(
    "ts_data", ["ttm-r1", "ttm-1024-96-r1", "ttm-r2", "ttm-1024-96-r2", "ttm-1536-96-r2"], indirect=True
)
def test_forecast_inference_decimal_freq(ts_data):
    test_data, params = ts_data

    prediction_length = params["prediction_length"]

    model_id = params["model_id"]
    model_id_path: str = model_id

    id_columns = params["id_columns"]

    # test single with integer as timestamp
    test_data_ = test_data[test_data[id_columns[0]] == "a"].copy()
    msg = {
        "model_id": model_id_path,
        "parameters": {
            # "prediction_length": params["prediction_length"],
        },
        "schema": {
            "timestamp_column": "date",
            "id_columns": params["id_columns"],
            "target_columns": params["target_columns"],
            "freq": "3600.0s",
        },
        "data": encode_data(test_data_, "date"),
        "future_data": {},
    }

    df_out, _ = get_inference_response(msg)
    assert len(df_out) == 1
    assert df_out[0].shape[0] == prediction_length


@pytest.mark.parametrize(
    "ts_data",
    [
        "ttm-r2-etth-finetuned",
    ],
    indirect=True,
)
def test_finetuned_model_inference(ts_data):
    test_data, params = ts_data
    id_columns = params["id_columns"]
    model_id = params["model_id"]

    prediction_length = 96

    # test single
    test_data_ = test_data[test_data[id_columns[0]] == "a"].copy()
    encoded_data = encode_data(test_data_, params["timestamp_column"])

    msg = {
        "model_id": model_id,
        "parameters": {
            # "prediction_length": params["prediction_length"],
        },
        "schema": {
            "timestamp_column": params["timestamp_column"],
            "id_columns": params["id_columns"],
            "target_columns": params["target_columns"],
        },
        "data": encoded_data,
        "future_data": {},
    }

    out = get_inference_response(msg)
    assert "Attempted to use a fine-tuned model with a different schema" in out.text

    # exact match of request and underlying schema
    # test_data_ = test_data_.drop(columns=params["timestamp_column"])
    msg = {
        "model_id": model_id,
        "parameters": {
            # "prediction_length": params["prediction_length"],
        },
        "schema": {
            "timestamp_column": params["timestamp_column"],
            "id_columns": params["id_columns"],
            "target_columns": ["OT"],
            "freq": "1h",
            "conditional_columns": [c for c in params["target_columns"] if c != "OT"],
        },
        "data": encoded_data,
        "future_data": {},
    }

    df_out, _ = get_inference_response(msg)
    assert len(df_out) == 1
    assert df_out[0].shape[0] == prediction_length

    # relaxed request
    msg = {
        "model_id": model_id,
        "parameters": {
            # "prediction_length": params["prediction_length"],
        },
        "schema": {
            "timestamp_column": params["timestamp_column"],
            "id_columns": params["id_columns"],
        },
        "data": encoded_data,
        "future_data": {},
    }

    df_out, _ = get_inference_response(msg)
    assert len(df_out) == 1
    assert df_out[0].shape[0] == prediction_length

    # relaxed request, but data issue (missing column):
    test_data_ = test_data[test_data[id_columns[0]] == "a"].copy()
    test_data_ = test_data_.drop(columns=params["target_columns"][0])
    encoded_data = encode_data(test_data_, params["timestamp_column"])

    msg = {
        "model_id": model_id,
        "parameters": {
            # "prediction_length": params["prediction_length"],
        },
        "schema": {
            "timestamp_column": params["timestamp_column"],
            "id_columns": params["id_columns"],
        },
        "data": encoded_data,
        "future_data": {},
    }

    out = get_inference_response(msg)
    assert "Attempted to use a fine-tuned model with data that does not match the saved schema" in out.text


@pytest.mark.parametrize(
    "ts_data",
    [
        "ttm-r2",
    ],
    indirect=True,
)
def test_improper_use_of_zero_shot_model_inference(ts_data):
    test_data, params = ts_data
    id_columns = params["id_columns"]
    model_id = params["model_id"]

    # conditional columns for non-conditional model
    test_data_ = test_data[test_data[id_columns[0]] == "a"].copy()
    encoded_data = encode_data(test_data_, params["timestamp_column"])

    msg = {
        "model_id": model_id,
        "parameters": {
            # "prediction_length": params["prediction_length"],
        },
        "schema": {
            "timestamp_column": params["timestamp_column"],
            "id_columns": params["id_columns"],
            "target_columns": ["OT"],
            "freq": "1h",
            "conditional_columns": [c for c in params["target_columns"] if c != "OT"],
        },
        "data": encoded_data,
        "future_data": {},
    }

    out = get_inference_response(msg)
    assert (
        "Unexpected parameter conditional_columns for a zero-shot model, please confirm you have the correct model_id and schema."
        in out.text
    )

    test_data_ = test_data[test_data[id_columns[0]] == "a"].copy()

    future_data = extend_time_series(
        select_by_index(test_data_, id_columns=params["id_columns"], start_index=-1),
        timestamp_column=params["timestamp_column"],
        grouping_columns=params["id_columns"],
        total_periods=25,
        freq="1h",
    )
    future_data = future_data.fillna(0)

    encoded_data = encode_data(test_data_, params["timestamp_column"])

    msg = {
        "model_id": model_id,
        "parameters": {
            # "prediction_length": params["prediction_length"],
        },
        "schema": {
            "timestamp_column": params["timestamp_column"],
            "id_columns": params["id_columns"],
            "target_columns": ["OT"],
            "freq": "1h",
        },
        "data": encoded_data,
        "future_data": encode_data(future_data, params["timestamp_column"]),
    }

    out = get_inference_response(msg)
    assert "Future data was provided, but the model does not support or require future exogenous." in out.text


@pytest.mark.parametrize(
    "ts_data",
    [
        "ibm/test-patchtst",
        "ibm/test-patchtsmixer",
    ],
    indirect=True,
)
def test_trained_model_inference(ts_data):
    test_data, params = ts_data
    id_columns = params["id_columns"]
    model_id = params["model_id"]

    prediction_length = 96

    # test single
    test_data_ = test_data[test_data[id_columns[0]] == "a"].copy()
    encoded_data = encode_data(test_data_, params["timestamp_column"])

    msg = {
        "model_id": model_id,
        "parameters": {
            # "prediction_length": params["prediction_length"],
        },
        "schema": {
            "timestamp_column": params["timestamp_column"],
            "id_columns": params["id_columns"],
            "target_columns": params["target_columns"],
        },
        "data": encoded_data,
        "future_data": {},
    }

    df_out, _ = get_inference_response(msg)
    assert len(df_out) == 1
    assert df_out[0].shape[0] == prediction_length



# Copyright contributors to the TSFM project
#
import logging
import os
import tempfile
from pathlib import Path

from .version import __version__, __version_tuple__


TSFM_PYTHON_LOGGING_LEVEL = os.getenv("TSFM_PYTHON_LOGGING_LEVEL", "INFO")

LevelNamesMapping = {
    "INFO": logging.INFO,
    "WARN": logging.WARN,
    "WARNING": logging.WARNING,
    "ERROR": logging.ERROR,
    "CRITICAL": logging.CRITICAL,
    "DEBUG": logging.DEBUG,
    "FATAL": logging.FATAL,
}

TSFM_PYTHON_LOGGING_LEVEL = (
    logging.getLevelNamesMapping()[TSFM_PYTHON_LOGGING_LEVEL]
    if hasattr(logging, "getLevelNamesMapping")
    else LevelNamesMapping[TSFM_PYTHON_LOGGING_LEVEL]
)
TSFM_PYTHON_LOGGING_FORMAT = os.getenv(
    "TSFM_PYTHON_LOGGING_FORMAT",
    "%(levelname)s:p-%(process)d:t-%(thread)d:%(filename)s:%(funcName)s:%(message)s",
)

TSFM_ALLOW_LOAD_FROM_HF_HUB = int(os.getenv("TSFM_ALLOW_LOAD_FROM_HF_HUB", "1")) == 1

# use TSFM_MODEL_DIR preferentially. If not set, use HF_HOME or the system tempdir if that's not set.
TSFM_MODEL_DIR: str = os.environ.get("TSFM_MODEL_DIR", os.environ.get("HF_HOME", tempfile.gettempdir()))

# basic checks
# make sure at least one of them is a valid directory
# make sure it's readable as well
_amodeldir_found = next(
    (
        adir
        for adir in (Path(p) for p in TSFM_MODEL_DIR.split(":"))
        if adir.exists() and adir.is_dir() and os.access(adir, os.R_OK)
    ),
    None,
)
if not _amodeldir_found and not TSFM_ALLOW_LOAD_FROM_HF_HUB:
    raise Exception(
        f"None of the values given in TSFM_MODEL_DIR {TSFM_MODEL_DIR} are an existing and readable directory."
    )



# Copyright contributors to the TSFM project
#

"""Tests conformal processor capabilities"""

import tempfile
from pathlib import Path

import numpy as np
import pytest
import torch

from tsfm_public.toolkit.conformal import (
    AdaptiveWeightedConformalScoreWrapper,
    NonconformityScores,
    PostHocGaussian,
    PostHocProbabilisticMethod,
    PostHocProbabilisticProcessor,
    WeightedConformalForecasterWrapper,
)


@pytest.mark.parametrize(
    "method", [PostHocProbabilisticMethod.CONFORMAL.value, PostHocProbabilisticMethod.GAUSSIAN.value]
)
def test_conformal_save_pretrained(method):
    # initial test to check that we save the ProbabbilisticProcessor as intended
    p = PostHocProbabilisticProcessor(method=method)

    with tempfile.TemporaryDirectory() as d:
        p.save_pretrained(d)
        p_new = PostHocProbabilisticProcessor.from_pretrained(d)
        assert Path(d).joinpath(PostHocProbabilisticProcessor.PROCESSOR_NAME).exists()

        assert p_new.method == method

        # to do: add checks that p and p_new are equivalent


def test_posthoc_probabilistic_processor():
    np.random.seed(42)

    # Parameters
    window_size = 10
    quantiles = [0.1, 0.5, 0.9]

    # Data
    n_samples = 30
    forecast_horizon = 5
    n_features = 3

    y_pred = np.random.randn(n_samples, forecast_horizon, n_features)
    y_gt = y_pred + np.random.normal(0, 0.5, size=y_pred.shape)

    y_cal_pred = y_pred[0:20]
    y_cal_gt = y_gt[0:20]
    y_test_pred = y_pred[-10:]
    # y_test_gt = y_gt[-10:]

    for method in [PostHocProbabilisticMethod.GAUSSIAN.value, PostHocProbabilisticMethod.CONFORMAL.value]:
        if method == PostHocProbabilisticMethod.CONFORMAL.value:
            nonconformity_score_list = [NonconformityScores.ABSOLUTE_ERROR.value, NonconformityScores.ERROR.value]
        else:
            nonconformity_score_list = [NonconformityScores.ABSOLUTE_ERROR.value]
        for nonconformity_score in nonconformity_score_list:
            # print(method,nonconformity_score )
            p = PostHocProbabilisticProcessor(
                window_size=window_size, quantiles=quantiles, nonconformity_score=nonconformity_score, method=method
            )
            p.train(y_cal_gt=y_cal_gt, y_cal_pred=y_cal_pred)
            y_test_prob_pred = p.predict(y_test_pred)

            ### ASSERTIONS ###

            ## Attributes checked correctly
            assert (
                p.method == method
            ), " method attribute wasnt assigned properly for method {} nonconformity score {}".format(
                method, nonconformity_score
            )
            assert (
                p.window_size == window_size
            ), " window_size attribute wasnt assigned properly for method {} nonconformity score {}".format(
                method, nonconformity_score
            )
            assert (
                p.quantiles == quantiles
            ), " quantiles attribute wasnt assigned properly for method {} nonconformity score {}".format(
                method, nonconformity_score
            )
            assert (
                p.nonconformity_score == nonconformity_score
            ), " nonconformity_score attribute wasnt assigned properly for method {} nonconformity score {}".format(
                method, nonconformity_score
            )

            assert (
                p.model.window_size == window_size
            ), " window_size attribute of processors model wasnt assigned properly for method {} nonconformity score {}".format(
                method, nonconformity_score
            )

            ## Check based on method
            if method == PostHocProbabilisticMethod.GAUSSIAN.value:
                assert isinstance(
                    p.model, PostHocGaussian
                ), "model is not an instance of PostHocGaussian for method {} nonconformity score {}".format(
                    method, nonconformity_score
                )

            if method == PostHocProbabilisticMethod.CONFORMAL.value:
                assert isinstance(
                    p.model, WeightedConformalForecasterWrapper
                ), "model is not an instance of WeightedConformalForecasterWrapper for method {} nonconformity score {}".format(
                    method, nonconformity_score
                )

            ## Prediction Output check
            assert isinstance(
                y_test_prob_pred, np.ndarray
            ), "Unexpected output type from predict(), it should be np array for method {} nonconformity score {}".format(
                method, nonconformity_score
            )
            assert y_test_prob_pred.shape == (
                y_test_pred.shape[0],
                forecast_horizon,
                n_features,
                len(quantiles),
            ), "Unexpected output shape from predict() for method {} nonconformity score {}".format(
                method, nonconformity_score
            )

            # Monotonicity check across quantiles
            assert np.all(
                y_test_prob_pred[..., 0] <= y_test_prob_pred[..., 1]
            ), "Quantile 0.1 is not <= 0.5 for method {} nonconformity score {}".format(method, nonconformity_score)
            assert np.all(
                y_test_prob_pred[..., 1] <= y_test_prob_pred[..., 2]
            ), "Quantile 0.5 is not <= 0.9 for method {} nonconformity score {}".format(method, nonconformity_score)


def test_posthoc_probabilistic_processor_online_update():
    """
    1. Generate Synthetic Data
    """
    np.random.seed(42)

    # Parameters
    window_size = 20
    quantiles = [0.1, 0.5, 0.9]

    # Data
    n_samples = 100
    n_cal = window_size
    forecast_horizon = 20
    n_features = 3
    sigma1 = 0.5
    sigma2 = 1.0
    transition_width = 10
    y_pred = np.random.randn(n_samples, forecast_horizon, n_features)

    # noise based on sigmoid transition from sigma1 to sigma2
    x = np.arange(n_samples)
    transition_center = n_samples // 2
    sigmoid = 1 / (1 + np.exp(-(x - transition_center) / (transition_width / 10)))
    sigma_values = sigma1 + (sigma2 - sigma1) * sigmoid
    sigma_expanded = sigma_values[:, None, None]  # shape: (n_samples, 1, 1)
    noise = np.random.normal(0, sigma_expanded, size=y_pred.shape)

    # ground truth
    y_gt = y_pred + noise

    y_cal_pred = y_pred[0:n_cal]
    y_cal_gt = y_gt[0:n_cal]

    y_test_pred = y_pred[n_cal:]
    y_test_gt = y_gt[n_cal:]

    """
    2. Run Methods with updates
    """
    nonconformity_score_list = [NonconformityScores.ABSOLUTE_ERROR.value]
    method_list = [PostHocProbabilisticMethod.CONFORMAL.value, PostHocProbabilisticMethod.GAUSSIAN.value]
    for method in method_list:
        for nonconformity_score in nonconformity_score_list:
            p = PostHocProbabilisticProcessor(
                window_size=window_size, quantiles=quantiles, nonconformity_score=nonconformity_score, method=method
            )

            ### 2. Fit
            p.train(y_cal_gt=y_cal_gt, y_cal_pred=y_cal_pred)

            ### 3. Predict Test (OFFLINE)
            y_test_prob_pred = p.predict(y_test_pred)

            ### 4. Predict Test with update (ONLINE)
            y_test_prob_pred_online = []
            for i in range(y_test_pred.shape[0]):
                # 4.1 Predict
                y_test_prob_pred_online.append(p.predict(y_test_pred[i : i + 1, ...]))
                # 4.2 Update (assume we observed the sample)
                p.update(y_gt=y_test_gt[i : i + 1, ...], y_pred=y_test_pred[i : i + 1, ...])

            y_test_prob_pred_online = np.stack(y_test_prob_pred_online)
            y_test_prob_pred_online = y_test_prob_pred_online.reshape(
                -1,
                y_test_prob_pred_online.shape[-3],
                y_test_prob_pred_online.shape[-2],
                y_test_prob_pred_online.shape[-1],
            )
            """
            Evaluation
            """
            coverage_online = []
            coverage_offline = []
            for q in range(len(quantiles)):
                covered_online = y_test_gt <= y_test_prob_pred_online[..., q]
                covered_offline = y_test_gt <= y_test_prob_pred[..., q]
                coverage_online.append(covered_online.mean(axis=(1, 2)))
                coverage_offline.append(covered_offline.mean(axis=(1, 2)))
            coverage_online = np.array(coverage_online).transpose()
            coverage_offline = np.array(coverage_offline).transpose()

            # coverage error
            error_online = np.abs(coverage_online - quantiles)
            error_offline = np.abs(coverage_offline - quantiles)

            mean_error_online = error_online.mean()
            mean_error_offline = error_offline.mean()

            # print(f"Mean Coverage Error (Online):  {mean_error_online:.4f}")
            # print(f"Mean Coverage Error (Offline): {mean_error_offline:.4f}")

            ### ASSERTIONS ###

            ## Prediction Output check
            assert isinstance(
                y_test_prob_pred, np.ndarray
            ), "Unexpected output type from predict(), it should be np array for method {} nonconformity score {}".format(
                method, nonconformity_score
            )
            assert y_test_prob_pred.shape == (
                y_test_pred.shape[0],
                forecast_horizon,
                n_features,
                len(quantiles),
            ), "Unexpected output shape from predict() for method {} nonconformity score {}".format(
                method, nonconformity_score
            )

            assert isinstance(
                y_test_prob_pred_online, np.ndarray
            ), "Unexpected output type from predict(), it should be np array for method {} nonconformity score {}".format(
                method, nonconformity_score
            )
            assert y_test_prob_pred_online.shape == (
                y_test_pred.shape[0],
                forecast_horizon,
                n_features,
                len(quantiles),
            ), "Unexpected output shape from predict() for method {} nonconformity score {}".format(
                method, nonconformity_score
            )

            ## Expected Behaviour of an online approach
            assert (
                mean_error_online <= mean_error_offline
            ), "Mean coverage error of online method should smaller that offline for method {} nonconformity score {}".format(
                method, nonconformity_score
            )


def test_posthoc_probabilistic_processor_outlier_score():
    """
    1. Generate Synthetic Data
    """
    np.random.seed(42)
    n = 250
    alarm_rate = 0.01
    window_size_min = int(np.ceil(1 / alarm_rate))
    size = int((n - window_size_min) * alarm_rate)

    t = np.linspace(0, 1, n)
    trend = 3 * t + 2 * t**2 - t**3
    baseline = 0.5 * np.sin(2 * np.pi * 5 * t)
    sigma = 0.2
    noise = np.random.normal(0, sigma, size=n)

    # Generate Signal
    signal = trend + baseline + noise

    # Inject only point anomalies
    anomaly_indices = np.random.choice(np.arange(window_size_min, n), size=size, replace=False)
    anomaly_magnitudes = 1.5
    anomaly_signs = np.random.choice([-1, 1], size=size)
    signal[anomaly_indices] += anomaly_magnitudes * anomaly_signs

    # Labels
    labels = np.zeros(n)
    labels[anomaly_indices] = 1

    # Naive Forecaster
    horizon = 1
    n_forecast = len(signal) - horizon

    # Create naive predictions
    y_pred_naive = np.tile(signal[:n_forecast], (horizon, 1)).T  # (n_forecast, horizon)
    y_pred_naive = y_pred_naive[..., np.newaxis]  # added feature dimension

    # Ground truth for evaluation
    y_true = np.array([signal[t + 1 : t + 1 + horizon] for t in range(n_forecast)])  # (n_forecast, horizon)
    y_true = y_true[..., np.newaxis]  # added feature dimension

    # Cal/Test Splits
    window_size = window_size_min
    y_cal_gt = y_true[0:window_size]
    y_cal_pred = y_pred_naive[0:window_size]
    y_test_pred = y_pred_naive[window_size:]
    y_test_gt = y_true[window_size:]

    quantiles = [alarm_rate / 2, 1 - alarm_rate / 2]

    """
    2. Run Method outlier score
    """
    nonconformity_score_list = [NonconformityScores.ABSOLUTE_ERROR.value]
    method_list = [PostHocProbabilisticMethod.CONFORMAL.value]
    for method in method_list:
        for nonconformity_score in nonconformity_score_list:
            p = PostHocProbabilisticProcessor(
                window_size=window_size, quantiles=quantiles, nonconformity_score=nonconformity_score, method=method
            )

            ### 2. Fit
            p.train(y_cal_gt=y_cal_gt, y_cal_pred=y_cal_pred)

            for aggregation in [0, "mean", "median"]:
                # print('AGGREGATION ', aggregation)
                ### 3. Outlier
                output_outlier = p.outlier_score(
                    y_pred=y_test_pred, y_gt=y_test_gt, significance=alarm_rate, aggregation=aggregation
                )

                labels_test = labels[-output_outlier.shape[0] :]
                labels_prediction_test = output_outlier[..., 1].flatten()
                p_value_scores_test = output_outlier[..., 0].flatten()

                ### ASSERTIONS ###

                ## Prediction Output check
                assert isinstance(
                    output_outlier, np.ndarray
                ), "Unexpected output type from outlier_score(), it should be np array for method {} nonconformity score {} on aggregation {}".format(
                    method, nonconformity_score, aggregation
                )
                assert (
                    output_outlier.shape
                    == (
                        y_test_pred.shape[0],
                        y_test_pred.shape[2],
                        2,
                    )
                ), "Unexpected output shape from predict() for method {} nonconformity score {} on aggregation {}".format(
                    method, nonconformity_score, aggregation
                )

                ### Expected Behaviour of an outlier approach ###

                # highest p-value for predicted outliers is below the alarm rate
                assert (
                    np.max(p_value_scores_test[labels_prediction_test == 1]) <= alarm_rate
                ), "Max p value among predicted outliers exceeds alarm rate for method {} nonconformity score {} on aggregation {}".format(
                    method, nonconformity_score, aggregation
                )

                # false positive rate is below the alarm rate
                assert (
                    np.mean(labels_prediction_test[labels_test == 0]) <= alarm_rate
                ), "False positive rate exceeds alarm rate for method {} nonconformity score {} on aggregation {}".format(
                    method, nonconformity_score, aggregation
                )

                # true positive rate is at least (1 - alarm_rate)
                assert (
                    np.mean(labels_prediction_test[labels_test == 1]) >= 1 - alarm_rate
                ), "True positive rate is lower than expectedfor method {} nonconformity score {} on aggregation {}".format(
                    method, nonconformity_score, aggregation
                )


def test_adaptive_conformal_wrapper():
    np.random.seed(42)
    torch.manual_seed(42)

    n_samples = 100
    std_dev = 0.05
    window_size = n_samples + 0
    false_alarm = 0.1
    window_size_critical = np.ceil(1 / false_alarm)

    """
    1. Synthetic Data Generation (scores)
    """
    error_drift = np.random.normal(0, std_dev, n_samples)
    error_scores = np.abs(error_drift)

    ### Test/Cal
    error_scores_cal = error_scores[0 : int(window_size_critical)]
    error_scores_test = error_scores[int(window_size_critical) :]

    """
    2. Initialization
    """
    n_batch_update = int(np.ceil(1 / false_alarm - 1) + 1)
    weighting_params = {
        "lr": 0.05,
        "n_batch_update": n_batch_update,
        "stride": n_batch_update,
        "epochs": 1,
        "conformal_weights_update": False,
    }

    TSAD_class = AdaptiveWeightedConformalScoreWrapper(
        false_alarm=false_alarm, window_size=window_size, weighting_params=weighting_params
    )

    ## Assertions Initialization
    assert (
        TSAD_class.false_alarm == false_alarm
    ), f"Expected false_alarm={false_alarm}, but got {TSAD_class.false_alarm}"
    assert (
        TSAD_class.window_size == window_size
    ), f"Expected window_size={window_size}, but got {TSAD_class.window_size}"
    for key in weighting_params.keys():
        assert key in TSAD_class.weighting_params, f"Key '{key}' not found in TSAD_class.weighting_params"
        assert (
            TSAD_class.weighting_params[key] == weighting_params[key]
        ), f"Mismatch for key '{key}' in TSAD_class.weighting_params : expected {weighting_params[key]}, got {TSAD_class.weighting_params[key]}"

    """
    3. Fit (its essentially init with calibration)
    """
    beta_cal = TSAD_class.fit(error_scores_cal)

    ## Assertions Fit
    assert (
        TSAD_class.weights_parameters.shape[0] == window_size
    ), f"Expected weights_parameters of length {window_size}, got {TSAD_class.weights_parameters.shape[0]}"
    assert (
        TSAD_class.weights_parameters.sum().item() == error_scores_cal.shape[0]
    ), f"Expected weights sum = {error_scores_cal.shape[0]}, got {TSAD_class.weights_parameters.sum().item()}"
    assert (
        TSAD_class.cal_scores.shape[0] == error_scores_cal.shape[0]
    ), f"Expected cal_scores.shape[0] = {error_scores_cal.shape[0]}, got {TSAD_class.cal_scores.shape[0]}"
    assert (
        beta_cal.shape[0] == error_scores_cal.shape[0]
    ), f"Expected beta_cal.shape[0] = {error_scores_cal.shape[0]}, got {beta_cal.shape[0]}"

    """
    4. Predict With the adaptive update of the weights
    """
    beta_test = TSAD_class.predict(error_scores_test)
    assert (
        beta_test.shape[0] == error_scores_test.shape[0]
    ), f"Expected beta_test.shape[0] = {error_scores_test.shape[0]}, got {beta_test.shape[0]}"
    # all scores come from the same distribution so weight norm should increase (all past values should be weighted)
    assert (
        TSAD_class.weights_parameters.sum().item() >= error_scores_cal.shape[0]
    ), f"Expected weights sum >= {error_scores_cal.shape[0]}, got {TSAD_class.weights_parameters.sum().item()}"


def test_forecast_horizon_aggregation():
    nonconformity_score = NonconformityScores.ABSOLUTE_ERROR.value
    method = PostHocProbabilisticMethod.CONFORMAL.value
    window_size = 100
    quantiles = [0.5]
    p = PostHocProbabilisticProcessor(
        window_size=window_size, quantiles=quantiles, nonconformity_score=nonconformity_score, method=method
    )
    ### Test 2
    outliers_scores_p = np.array([[0.1, 0.01, 0.001], [0.1, 0.01, 0.005], [0.1, 0.05, 0.0001], [0.5, 0.01, 0.0001]])[
        :, :, np.newaxis
    ]  # feature 0
    outliers_scores_ix = np.array([[1, 2, 3], [2, 3, 4], [3, 4, 5], [4, 5, 6]])[:, :, np.newaxis]  # feature 1
    outliers_scores = np.concatenate([outliers_scores_p, outliers_scores_ix], axis=-1)

    expected_aggregation_ix = np.array([[1], [2], [3], [4]])
    expected_aggregation = {}
    expected_aggregation["min"] = np.concatenate(
        [np.array([[0.1], [0.01], [0.001], [0.005]]), expected_aggregation_ix], axis=-1
    )
    expected_aggregation["max"] = np.concatenate(
        [np.array([[0.1], [0.1], [0.1], [0.5]]), expected_aggregation_ix], axis=-1
    )
    expected_aggregation["median"] = np.concatenate(
        [
            np.array(
                [[0.1], [np.median([0.1, 0.01])], [np.median([0.1, 0.01, 0.001])], [np.median([0.5, 0.05, 0.005])]]
            ),
            expected_aggregation_ix,
        ],
        axis=-1,
    )
    expected_aggregation["mean"] = np.concatenate(
        [
            np.array([[0.1], [np.mean([0.1, 0.01])], [np.mean([0.1, 0.01, 0.001])], [np.mean([0.5, 0.05, 0.005])]]),
            expected_aggregation_ix,
        ],
        axis=-1,
    )
    expected_aggregation[0] = np.concatenate(
        [np.array([[0.1], [0.1], [0.1], [0.5]]), expected_aggregation_ix], axis=-1
    )
    expected_aggregation[1] = np.concatenate(
        [np.array([[0.1], [0.01], [0.01], [0.05]]), expected_aggregation_ix], axis=-1
    )
    expected_aggregation[2] = np.concatenate(
        [np.array([[0.1], [0.01], [0.001], [0.005]]), expected_aggregation_ix], axis=-1
    )
    for aggregation in ["min", "mean", "max", "median", 0, 1, 2]:
        outliers_aggregated = p.forecast_horizon_aggregation(outliers_scores, aggregation=aggregation)
        # print(aggregation)
        # print(outliers_aggregated)
        # print(expected_aggregation[aggregation])
        # print()
        assert (
            outliers_aggregated.shape == (outliers_scores.shape[0], outliers_scores.shape[-1])
        ), f"forecast_horizon_aggregation should provide an output of shape {(outliers_scores.shape[0],outliers_scores.shape[-1])}"
        assert (
            np.mean(np.abs(outliers_aggregated - expected_aggregation[aggregation])) == 0
        ), f"Expected forecast_horizon_aggregation for aggregation {aggregation} did not match the expected values"

    ### Test 3
    window_size = 20
    quantiles = [0.1]
    p = PostHocProbabilisticProcessor(
        window_size=window_size, quantiles=quantiles, nonconformity_score=nonconformity_score, method=method
    )

    y_cal_gt = np.linspace(0, 2, 21)[1:-1] - 1
    y_cal_gt = y_cal_gt[:, np.newaxis, np.newaxis] * np.ones([1, 3, 4])
    y_cal_pred = np.zeros_like(y_cal_gt)

    ### 2. Fit
    p.train(y_cal_gt=y_cal_gt, y_cal_pred=y_cal_pred)

    # N = 0 shouldn't be an outlier
    # N=1 should be an outlier if considering h=1, but not h=0
    # N=2 shouldn't be an outlier
    # N=3 should be an outlier if considering h=2, but not h=1 or h=0

    y_test_gt = np.array(
        [
            [0.7, 0.9, 0.8],  # not an outlier
            [0.5, 0.7, 0.9],  # outlier for h=1 and h=2 (cause no pred of h=2 is available so using h=1)
            [0.5, 0.7, 0.9],  # No outlier
            [0.5, 1.0, 1.0],
        ]
    )[..., np.newaxis] * np.ones([1, 1, 4])  # outlier for h=2
    y_test_pred = np.zeros_like(y_test_gt)
    significance = 0.1

    outlier_gt = {}
    outlier_gt[0] = 0
    outlier_gt[1] = 1 * y_test_gt.shape[-1]
    outlier_gt[2] = 2 * y_test_gt.shape[-1]
    outlier_gt["max"] = 0
    outlier_gt["min"] = 2 * y_test_gt.shape[-1]

    ### Outliers count
    for aggregation in outlier_gt.keys():
        output_outlier = p.outlier_score(
            y_pred=y_test_pred, y_gt=y_test_gt, significance=significance, aggregation=aggregation
        )
        assert (
            np.sum(output_outlier[..., 1]) == outlier_gt[aggregation]
        ), f"Number of outliers for aggregation {aggregation} did not match the expected values"


if __name__ == "__main__":
    # test_posthoc_probabilistic_processor_outlier_score()
    # test_adaptive_conformal_wrapper()
    test_forecast_horizon_aggregation()
